#ifndef  VEC_UTILITY_H
  #include "vec_utility.H"
#endif
#include "protein.H"

/* 
   I will use simple distance-dependent dielectric as described in p. 169 of Leach. 
   In the future, may use sigmoidal shape function, again same page.
   Also, the function is not continuous in the first derivative, it is lower and upper bounded.
*/
float coulombic(const Point p1, float q1, const Point p2, float q2) {

  float retval, distSqr;

  distSqr  = (p1[0] - p2[0]) * (p1[0] - p2[0]);
  distSqr += (p1[1] - p2[1]) * (p1[1] - p2[1]);
  distSqr += (p1[2] - p2[2]) * (p1[2] - p2[2]);
  retval = 332 * q1 * q2 / (distSqr);
  if (retval < -5)
    retval = -5;
  else if (retval > 5)
    retval = 5;
  
  return retval;
}


void Protein::setHydrophobicity()
{
  const int  numHResidues                = 9;
  const char hydrophobics[numHResidues]  = 
  {'A','C','V', 'F', 'M', 'I', 'W', 'L', 'Y'};

  for (char ch = 'A'; ch <= 'Z'; ch ++) 
    {
      hydrophobic[ch] = false;
    }
  for (int i = 0; i < numHResidues; i++) 
    {
      hydrophobic[hydrophobics[i]] = true;
      
    }
}

#if 0
/* AAcode assumed to be uppercase */
bool hydrophobic(char AAcode) {

  const char hydrophobics[9]  = {'A','C','V', 'F', 'M', 'I', 'W', 'L', 'Y'};
  const char polars[11]       = {'D','E','K','R', 'P','S','T','H','N','Q','G'};
  int i;

  for (i = 0; i< 9; i++) {
    if (AAcode == hydrophobics[i])
      return true;
  }

  for (i = 0; i< 11; i++) {
    if (AAcode == polars[i])
      return false;
  }
  
  cout << "Could not find : " << AAcode << " among the possible AA codes... " << endl;
  exit(0);
  return false;
}
#endif
float exclusionEnergy_Calpha (float distance_squared) {
  // distance_squared is the cAlpha to  cAlpha distance squared.
  //  constants chosen so that error < 1E-4.

  const float d_eff    = 3.6;
  const float dw       = 0.1;
  const float C        = 5.0;
  float denum, distance;
    
  if (distance_squared > 20.5) { 
    return 0.0;
  }
  else if (distance_squared < 6.76) {
    return C;
  }
  else {
    distance = sqrt(distance_squared);
    denum    = 1 + exp((distance - d_eff)/dw);
    return (C/denum);
  }
}

/* discontinuous, returns very high energy if the distance between atoms < 1A. 
   otherwise, same as exclusionEnergyCentroid.
   to be used with coulombic forces, which may favor collisions..
*/
float exclusionEnergy_Centroid_withCoulombic(float distance_squared) {
  
  const float d_eff    = 3.2;
  const float dw       = 0.1;
  const float C        = 5.0;
  const float VERY_HIGH_VALUE = 100000.0;
  float denum, distance;
    
  if (distance_squared > 17) { //chosen so that error < 1E-4. 
    return 0.0;
  }
  else if (distance_squared < 4.84) { 
    if (distance_squared < 1) {
      return VERY_HIGH_VALUE;
    }
    else 
      return C;
  }
  else {
    distance = sqrt(distance_squared);
    denum    = 1 + exp((distance - d_eff)/dw);
    return (C/denum);
  }
}


float exclusionEnergy_Centroid (float distance_squared) {
  // distance_squared is the centroid to  centroid distance squared.
  //  constants chosen so that error < 1E-4.

  const float d_eff    = 3.2;
  const float dw       = 0.1;
  const float C        = 5.0;
  float denum, distance;
    
  if (distance_squared > 17) { //chosen so that error < 1E-4. 
    return 0.0;
  }
  else if (distance_squared < 4.84) { 
    return C;
  }
  else {
    distance = sqrt(distance_squared);
    denum    = 1 + exp((distance - d_eff)/dw);
    return (C/denum);
  }
}


float HP_Factor_Altered(float distance_sqr) {
  /* distance_sqr is the square of the centroid2centroid or cAlpha to cAlpha distance */

  const float dO = 6.5;
  const float dt = 2.5;
  float denum;

  if (distance_sqr > 872.0) { // this number selected so as to have the computed value < 1E-4
    return 0.0;
  }
  else {
    float d = sqrt(distance_sqr);
    denum = 1.0 + exp((d - dO)/dt);
    return (-1 / denum);
  }
}


