/* 
   3/21 : code changed to adapt the removal of the first and/or last 
   loop vectors in some proteins.
   This program generates side chain data that will be used to compute 
   the energy of the protein in protein.C
*/
//#include <fstream.h>
#include "vec_utility.H"
#include <assert.h>
#include <string.h>
#include <stdlib.h>
#include <vector>
#include "protein.H"
#include "sse.H"
#include "vec_utility.H"
#include <stdio.h>
#include <string>

extern void extStr(char* out, const char* in, int st, int end);
extern void standardizeNativeStructure (Protein * standardP, 
					Protein * originalP);
extern void findBaseAndUnitVecsForTheWholeSse(float base[], float unitX[],
					      float unitY[], float unitZ[], 
					      const Protein *myProtein, 
					      int currentSseElementIndex);


void computeTransformationMatrix   (mat4  transformMat, 
				    const Protein * standardProt, 
				    const Protein * myProtein);

int  findSseElementIndex           (const Protein * protein, int AAindex);
void printPairwiseCAlphaDistances  (Protein * myProtein);

const int SSE_PART_OF_REMOVED_LOOP = -5;

/* extracts the coordinates of the N, c-alpha and C atoms from the lines */
void extractAtomCoordinates(float point[], const char * line) 
{
	
  char record[100];
  

  extStr(record, line, 31,38);
  point[0] = (float) (atof(record));
  extStr(record, line, 39,46);
  point[1] = (float) (atof(record));
  extStr(record, line, 47,54);
  point[2] = (float) (atof(record));
}

void updateCentroid (float * sideChainWeight, float centroid[],
		     const char elementCode, const float newPoint[]) 
{
  /*
    map utilization example
    map<char, float> atomWeights;
    atomWeights['O'] = 16.0;atomWeights['S'] = 32.0;
    atomWeights['C'] = 12.0;atomWeights['N'] = 14.0; 
    atomWeights['H'] = 1.0;
  */
  float weight;
  
  switch(elementCode) {
  case 'O' : {
    weight = 16.0;
    break;
  } 
  case 'S' : { 
    weight = 32.0;
    break;
  } 
  case 'C' : { 
    weight = 12.0;
    break;
  }
  case 'N' : { 
    weight = 14.0;
    break;
  }
  case 'H' : { 
    weight = 1.0;
    break;
  }
  default : { 
    cout << "Could not find weight for : " << elementCode << endl;
    exit(1);
  }
  
  } // switch
  
  centroid[0]   += newPoint[0] * weight;
  centroid[1]   += newPoint[1] * weight;
  centroid[2]   += newPoint[2] * weight;

  (*sideChainWeight) += weight;
  
}

// point_N, point_O, point_C are also to be returned.
// added point_CA as output for finding rel. position of C-alpha atom.

void getFirstFourAtoms(Point point_CA, const char * firstLine, 
		       FILE * infile, 
		       Point point_N,  Point point_Cprime, Point point_O) 
{
  char  line[200], record[50];

  extStr(record, firstLine,18 ,20);

  // getting N,Ca, C and O coordinates
  extractAtomCoordinates(point_N,firstLine);
  fgets(line,200,infile);
  extStr(record, line, 14, 15);
  if (strcmp(record, "CA")) 
    {
      cout << "Record = " << record << " line = " << line << endl;
    }
  assert(strcmp(record, "CA") == 0);
  extractAtomCoordinates(point_CA,line);
  fgets(line,200,infile);
  extStr(record, line, 14, 15);
  assert(strcmp(record, "C ") == 0);
  extractAtomCoordinates(point_Cprime,line);
  fgets(line,200,infile);
  extStr(record, line, 14, 15);
  assert(strcmp(record, "O ") == 0);
  extractAtomCoordinates(point_O,line);
}



// cylindirical coordinates are r,phi and z, accordingly.
void cartesian2Cylindrical(float cylCoordinates[], const vec4 relCoordinates) 
{
  const float epsilon = 1E-3;

  if (!(fabs(relCoordinates[0] - 1.0) < epsilon)) 
    {
      cout << relCoordinates[0] << " " << relCoordinates[1] << " " << relCoordinates[2] << endl;
      cout << cylCoordinates[0] << " " << cylCoordinates[1] << " " << cylCoordinates[2] << endl;
    }

  assert(fabs(relCoordinates[0] - 1.0) < epsilon);
  /* z */    cylCoordinates[2] = relCoordinates[3];
  /* r */    cylCoordinates[0] = sqrt(relCoordinates[1]*relCoordinates[1] + relCoordinates[2]*relCoordinates[2]);
  if ((relCoordinates[1] == 0.0) && (relCoordinates[2] == 0.0)) 
    {
      /* phi */   cylCoordinates[1] = 0;
    }
  else 
    {
      /* phi */   cylCoordinates[1]   = atan2(relCoordinates[2],relCoordinates[1]);
    }
}


// returns cylCoordinates, which is a vector of 4 elements.
// cylindirical coordinates are 1, r,phi and z, accordingly.
void cartesian2Cylindrical (vector<float>&  cylCoordinates, 
			    const vec4 relCoordinates) 
{
  const float epsilon = 1E-3;

  if (!(fabs(relCoordinates[0] - 1.0) < epsilon)) 
    {
      cout << relCoordinates[0] << " " << relCoordinates[1] << " " << relCoordinates[2] << endl;
      cout << cylCoordinates[0] << " " << cylCoordinates[1] << " " << cylCoordinates[2] << endl;
    }

  assert(fabs(relCoordinates[0] - 1.0) < epsilon);
  
  /* z */    cylCoordinates[3] = relCoordinates[3];
  /* r */    cylCoordinates[1] = sqrt(relCoordinates[1]*relCoordinates[1] + relCoordinates[2]*relCoordinates[2]);
  if ((relCoordinates[1] == 0.0) && (relCoordinates[2] == 0.0)) 
    {
      /* phi */   cylCoordinates[2] = 0;
    }
  else 
    {
      /* phi */   cylCoordinates[2] = atan2(relCoordinates[2],relCoordinates[1]);
    }
  cylCoordinates[0] = 1;
}


#if 0

no longer utilized. instead, use the function from protein.C w/ the same name that also computes unitX and unitY.

/* difference from findBaseAndUnitVecs : the base and unit vectors don't change according to relAAindex, only depend on which SSE we are at. Also, the unit vector has unit norm, whereas in the other function it is for one AA but it does not necessarily have unit norm.
   currentSseElementIndex == 0 .. numSseElements -1.
*/
void findBaseAndUnitVecsForTheWholeSse(float base[], float unit[], Protein *myProtein, int currentSseElementIndex ) 
{
  
  VecCopy(base, (myProtein->getSseElement(currentSseElementIndex))->getBasePoint());
  VecSub (unit, (myProtein->getSseElement(currentSseElementIndex))->getEndPoint(), (myProtein->getSseElement(currentSseElementIndex))->getBasePoint());
  VecNormalize(unit);
  
   cout << "base and unit vectors for the AA in SSe # " << currentSseElementIndex << endl;
   for (int i = 0; i< 3; i++) {
     cout << base[i] << " "  << unit[i] << endl;
   }
  
}
#endif


/* 
   finds the coordinates absCoordinates in the coordinate frame of the
   corresponding vector on the protein, and then transforms these coordinates
   into cylindirical coordinates. Returns in cylCoordinates.
   
   if the corresponding AA is on a loop, returns -1 instead.

   assumes that cylCoordinates is already allocated.

   absCoords and cylCoordinates are 4 element vectors, with the first
   element being equal to 1.  (homogenous coordinates)
   AAindex = 1..numResidues.

   this routine can be greatly enhanced, by computing unit vector only once,
   for the whole SSE vector. 
   For this, the call to this function should perhaps include an extra 
   parameter for this variable, 
   which would not be recomputed in the same SSE vector.
*/
void computeCylindiricalCoordsFromAbsoluteCoords(vector<float> &  
						 cylCoordinates,  
						 const vector<float> & 
						 absCoords, 
						 Protein * myProtein, 
						 const int AAindex) 
{
  int currentSseElementIndex = 0; // 0.. # sse Elements - 1
  float base[3],                  // base of the vector corresponding to
                                  // AAindex
    unitZ[3],                     // unit vec. corresponding to axis of
                                  // the SSE vector
    unitY[3],    
    unitX[3];
  mat4 T;
  vec4 absCoordinates, relCoordinates;
  
  assert (( AAindex >= 1) && ( AAindex <= Protein::getNumAAs()));
  //  cout << absCoords[0] << endl;
  assert (fabs(absCoords[0] - 1) < 1E-3);

  currentSseElementIndex = findSseElementIndex(myProtein, AAindex);
  
  if (currentSseElementIndex == SSE_PART_OF_REMOVED_LOOP) 
    {
      for (int i = 0; i < 4; i++)
	{
	  cylCoordinates[i] =  -1;
	}
      return;
    }
  
  if (currentSseElementIndex != -1) 
    {
      findBaseAndUnitVecsForTheWholeSse (base, unitX, unitY, unitZ, myProtein,
					 currentSseElementIndex);
    }
  else 
    {
      cout << "Sorry, could not find the corresponding sse element. " << endl;
      assert(0);
    }

  T[0][1] = 0;        T[0][2] = 0;        T[0][3] = 0; 
  T[1][1] = unitX[0]; T[1][2] = unitX[1]; T[1][3] = unitX[2]; 
  T[2][1] = unitY[0]; T[2][2] = unitY[1]; T[2][3] = unitY[2]; 
  T[3][1] = unitZ[0]; T[3][2] = unitZ[1]; T[3][3] = unitZ[2]; 
  

  T[0][0] =     1.0; 
  T[1][0] = - ( T[1][1] * base[0] + T[1][2] * base[1] + T[1][3] * base[2]);
  T[2][0] = - ( T[2][1] * base[0] + T[2][2] * base[1] + T[2][3] * base[2]);
  T[3][0] = - ( T[3][1] * base[0] + T[3][2] * base[1] + T[3][3] * base[2]);
  
  for (int i = 0; i < 4; i++)
    {
      absCoordinates[i] = absCoords[i]; 
    }
  
  /* 
     finds the coords of centroid in the new cartesian coord system 
     centered at base and w/ unit vecs given by unitX, unitY, unitZ. 
  */
  /*
  cout << "the sse is : " << endl;
  cout << "base : " <<  base[0] << " " << base[1] << " " << base[2] << endl;
  cout << "unitX : " << unitX[0] << " " <<  unitX[1] << " "  << unitX[2] << endl;
  cout << "unitY : " << unitY[0] << " " <<  unitY[1] << " "  << unitY[2] << endl;
  cout << "unitZ : " << unitZ[0] << " " <<  unitZ[1] << " "  << unitZ[2] << endl;
  cout << "sse index : " << currentSseElementIndex << endl;
  cout << "AAindex = " << AAindex << endl;
  cout << "original CA coords : " << absCoordinates[0] << " ";
  cout << absCoordinates[1] << " " << absCoordinates[2] << " " ;
  cout << absCoordinates[3] << endl;
  */
  applyMatToVec4(relCoordinates, T, absCoordinates);

  /*
  cout << "CA coords in the ref. frame of the sse: " << endl;
  cout << relCoordinates[0] << " " << relCoordinates[1] << " ";
  cout << relCoordinates[2] << " " << relCoordinates[3] << endl;
  */
  cartesian2Cylindrical(cylCoordinates, relCoordinates);

  /*
  cout << "cyl coords obtained: ";
  cout << cylCoordinates[0] << " " << cylCoordinates[1] << " " << cylCoordinates[2] << " " << cylCoordinates[3] << endl;
  cout << "exiting in computeCylindiricalCoordsFromAbsoluteCoords" << endl;
  */
  //  exit(0); 

}


#if 0
/* 
   AAindex = 1..numResidues.
   this routine can be greatly enhanced, by computing unit vector only once,
   for the whole SSE vector. 
   For this, the call to this function should perhaps include an extra 
   parameter for this variable, 
   which would not be recomputed in the same SSE vector.
*/
void computeCylindiricalCoordsFromAbsoluteCoords(float cylCoordinates[],  
						 const float centroid[], 
						 Protein * myProtein, 
						 const int AAindex) 
{
  
  int currentSseElementIndex = 0; // 0.. # sse Elements - 1
  float base[3], // base of the vector corresponding to AAindex
    unitZ[3],    // unit vec. corresponding to axis of the SSE vector
    unitY[3],    
    unitX[3];
  //const float epsilon = 1E-4;
  mat4 T;
  vec4 absCoordinates, relCoordinates;
  
  currentSseElementIndex = findSseElementIndex(myProtein, AAindex);
  
  if (currentSseElementIndex == SSE_PART_OF_REMOVED_LOOP) 
    {
      cylCoordinates[0] = cylCoordinates[1] = cylCoordinates[2] = -1;
      return;
    }
  
  if (currentSseElementIndex != -1) 
    {
      findBaseAndUnitVecsForTheWholeSse (base, unitX, unitY, unitZ, myProtein,
					 currentSseElementIndex);
    }
  else 
    {
      cout << "Sorry, could not find the corresponding sse element. " << endl;
      assert(0);
    }

      
#if 0
    //    findBaseAndUnitVecs(base, unitZ, myProtein, currentSseElementIndex,  AAindex - numAASoFar - 1);
    findBaseAndUnitVecsForTheWholeSse(base, unitZ, myProtein, currentSseElementIndex);
    assert(VecLength(unitZ) == 1.0);
    unitY[0] = -unitZ[1]; unitY[1] = unitZ[0]; unitY[2] = 0;
    VecNormalize(unitY);
    
    if (fabs(VecLength(unitY) - 1.0) > epsilon) {
	cout << "unitY = " << unitY[0] << " " << unitY[1] << " " << unitY[2] << endl;
	unitY[0] = 1.0; unitY[1] = unitY[2] = 0.0;
	cout << "unitY atbitrarily selected as (1,0,0), since unitZ is : " << unitZ[0] << " " << unitZ[1] << " " << unitZ[2] << endl;
    }

    VecCross(unitX, unitY, unitZ);
    assert (fabs(VecLength(unitX) - 1.0) < epsilon);
    cout << "computed unitX : " << unitX[0] << " " << unitX[1] << " " << unitX[2] << endl;
    cout << "computed unitY : " << unitY[0] << " " << unitY[1] << " " << unitY[2] << endl;
#endif
    
    T[0][1] = 0;        T[0][2] = 0;        T[0][3] = 0; 
    T[1][1] = unitX[0]; T[1][2] = unitX[1]; T[1][3] = unitX[2]; 
    T[2][1] = unitY[0]; T[2][2] = unitY[1]; T[2][3] = unitY[2]; 
    T[3][1] = unitZ[0]; T[3][2] = unitZ[1]; T[3][3] = unitZ[2]; 
    
    
    T[0][0] = 1.0; T[1][0] = - (T[1][1] * base[0] + T[1][2] * base[1] + T[1][3] * base[2]);
    T[2][0] = - ( T[2][1] * base[0] + T[2][2] * base[1] + T[2][3] * base[2]);
    T[3][0] = - ( T[3][1] * base[0] + T[3][2] * base[1] + T[3][3] * base[2]);
    
    absCoordinates[0] = 1;           absCoordinates[1] = centroid[0]; 
    absCoordinates[2] = centroid[1]; absCoordinates[3] = centroid[2]; 
    /* 
       finds the coords of centroid in the new cartesian coord system 
       centered at base and w/ unit vecs given by unitX, unitY, unitZ. 
    */
    applyMatToVec4       (relCoordinates, T, absCoordinates);
    cartesian2Cylindrical(cylCoordinates, relCoordinates);
}
#endif
#if 0

void computeOneCentroid (Point point_CA, Point centroid, FILE * infile, 
			 const char * firstLine, 
			 char * nextLine, Point point_N, Point point_Cprime, 
			 Point point_O) 
{

  float point[3];
  int number;     // AA number of the coming up atoms (after N, C-a and C')
  char line[200], record[50];
  int AAnumber;

  getFirstFourAtoms(point_CA, firstLine, infile, point_N, point_Cprime, 
		    point_O);
  
  /* 
     getting the rest of the atom coordinates 
  */
  
  extStr(record,firstLine,23,26);
  AAnumber = (int) atof(record); 
  centroid[0] = centroid[1] = centroid[2] = 0.0;
  
  //cout << "Computing the " << AAnumber << " th amino acid's centroid." << endl;
  float weightOfSideChain = 0.0;  // this is for debugging

  do 
    {
      fgets (line,200,infile);
      extStr(record,line,23,26);
      number = (int) atof(record);
      if (number != AAnumber) 
	{
	  assert(number == AAnumber + 1);
	  strcpy(nextLine, line);
	  if (weightOfSideChain == 0.0) 
	    {
	      VecCopy(centroid, point_CA);
	      cout << "There are no atoms in the sidechain for AA# " << AAnumber << " . Using C_alpha loc. instead." << endl;
	    }
	  else 
	    {
	      centroid[0] /= weightOfSideChain; 
	      centroid[1] /= weightOfSideChain; 
	      centroid[2] /= weightOfSideChain;
	    }
	}
      else 
	{
	  extStr(record, line, 1,3);
	  if (!strcmp(record, "TER")) 
	    {
	      cout << "Record = " << record << " breaking... " << endl;
	      strcpy(nextLine,line);
	      if (weightOfSideChain == 0.0) {
		VecCopy(centroid, point_CA);
		cout << 
		  "There are no atoms in the sidechain for AA# " << AAnumber << " . Using C_alpha loc. instead." << endl;
	      }
	      else 
		{
		  centroid[0] /= weightOfSideChain; 
		  centroid[1] /= weightOfSideChain; 
		  centroid[2] /= weightOfSideChain;
		}
	      break;
	    }
	  else 
	    {
	      extractAtomCoordinates(point,line);
	      //	      cout << "Extra atom code :" << line[13] << endl;
	      updateCentroid(&weightOfSideChain, centroid,line[13], point);
	      // netSideChainWeights[*SC_location_firstIndex]);       
	    }
	}
    } while (number == AAnumber);
  
  //  cout << "computed values  : " << centroid[0] << " " << centroid[1] << " " << centroid[2] << " for AA :" << AAnumber << endl;
}

#if 0
do not use until you make sure that this function works even if we clip off the first and the last vectors, in case they are loops.
/* AAindex =  1.. numResidues. */
int findRelAAIndex(int AAindex, const Protein *myProtein) {

    int numAASoFar = 0, // # AAs encountered till the current sse.
	currentSseElementIndex = 0; // 0 .. # sse Elements - 1
    bool found = false;
    
    assert((AAindex > 0) && (AAindex <= myProtein->getNumAAs()));
    while ((AAindex > numAASoFar) && (currentSseElementIndex < myProtein->getNumSseElements())) {
	
	if (AAindex <= (numAASoFar + ((myProtein->getSseElement(currentSseElementIndex))->getNumAAs()))) {
	    // found the corresponding sse Element
	    found = true;
	    break;
	}
	else {
	    numAASoFar +=  (myProtein->getSseElement(currentSseElementIndex))->getNumAAs();
	    currentSseElementIndex ++;
	}
    }
    if (found)
	return  (AAindex - numAASoFar - 1);
    else {
	cout << "findRelAAIndex could not find relAAindex for AAindex == " << AAindex << endl;
	exit(0);
    }
    return 0;
}
#endif
#endif
#if 0
/* 
   computes the given coords in the new coordinate system, and then 
   translates it to cylindirical coords.
*/

void writeCylCoordsInfo (ofstream * fout, vec4 point, mat4 transformMat, 
			 Protein * stProt, int AAindex)
{

  float cylCoordinates[3];
  vec4  XformedPoint;  // point in standardized coordinate system
  
  applyMatToVec4                              (XformedPoint, transformMat, 
					       point);
  
  computeCylindiricalCoordsFromAbsoluteCoords (cylCoordinates, 
					       &XformedPoint[1], stProt, 
					       AAindex);
  
  (*fout) << cylCoordinates[1] << " " << cylCoordinates[2] << " " << cylCoordinates[3] << endl;

}
#endif

/* 
   writes coords to the given filename
*/

void writeCoords2File (const char * filename, 
		       const vector<vector<float> >& coords)
{
  ofstream fout (filename);

  for (unsigned int i = 0; i < coords.size(); i++)
    {
      //      cout << coords[i][0] << endl;
      assert ((fabs(coords[i][0] - 1) < 1E-2) || (fabs(coords[i][0] + 1) < 1E-2));
      for (unsigned int j = 1; j < 4;  j++)
	{
	  fout << coords[i][j] << " ";
	}
      fout << endl;
    }
  cout << "check " << filename << endl;
}

/*
  just writes the files.
*/

void writeAllCoords2Files (const vector<vector<float> >& relNCoords, 
			   const vector<vector<float> >& relCACoords, 
			   const vector<vector<float> >& relCpCoords, 
			   const vector<vector<float> >& relHCoords, 
			   const vector<vector<float> >& relOCoords, 
			   const vector<vector<float> >& 
			   relCentroidCoords)
			   
{
  writeCoords2File ("cAlphaRelDict.txt",   relCACoords);
  writeCoords2File ("CprimeRelDict.txt",   relCpCoords);
  writeCoords2File ("centroidRelDict.txt", relCentroidCoords);
  //  writeCoords2File ("HRelDict.txt",        relHCoords);
  writeCoords2File ("ORelDict.txt",        relOCoords);
  writeCoords2File ("NRelDict.txt",        relNCoords);
}

/*
  given protein, coords, and transform matrix, computes rel. coords.
  first applies the transformation matrix to the point.
  then  computes its cylindirical coordinates in the reference frame
        of the standard protein.
*/

void computeRelCoords (vector<vector<float> > & relCoords,  
		       const vector<vector<float> > & coords,  
		       mat4      transformMat, 
		       Protein * standardProt)
{
  static bool firstCall = true;
  static vector<vector<float> > standardizedCoords;
  
  relCoords.resize          (coords.size());
  for (unsigned int i = 0; i < coords.size(); i++)
    {
      relCoords[i].resize (4);
    }
  
  if (firstCall) 
    {
      /* 
	 first allocate space 
      */
      standardizedCoords.resize (coords.size());

      for (unsigned int i = 0; i < coords.size(); i++)
	{
	  standardizedCoords[i].resize(4);
	}
      firstCall = false;
    }
  
  /* 
     now compute the standardized coords for each 
  */
  
  // standardized coords: coords wrt a protein whose first vectors basis
  // lies at the origin.
  
  assert (relCoords.size()          == coords.size());
  assert (standardizedCoords.size() == coords.size());
  
  for (unsigned int i = 0; i < coords.size(); i++)
    {
      applyMatToVec4 (standardizedCoords[i],  transformMat, coords[i]);
    }
  
  /* 
     now compute the cyl. coordinates from the standardized coords. 
  */
  for (unsigned int i = 0; i < coords.size(); i++)
    {
      computeCylindiricalCoordsFromAbsoluteCoords (relCoords[i], 
						   standardizedCoords[i], 
						   standardProt, 
						   i+1);
    }
}

/* 
   calls computeRelCoords for all atom types 
*/

void computeRelCoords4AllAtoms (vector<vector<float> > & relNCoords, 
				vector<vector<float> > & relCACoords, 
				vector<vector<float> > & relCpCoords, 
				vector<vector<float> > & relHCoords, 
				vector<vector<float> > & relOCoords, 
				vector<vector<float> > & relCentroidCoords,
				const vector<vector<float> > &  NCoords, 
				const vector<vector<float> > &  CACoords, 
				const vector<vector<float> > &  CpCoords,
				const vector<vector<float> > &  HCoords, 
				const vector<vector<float> > &  OCoords, 
				const vector<vector<float> > &  
				centroidCoords,
				mat4 transformMat, Protein * standardProt)
{
  computeRelCoords (relCACoords, CACoords, transformMat, standardProt);
  computeRelCoords (relNCoords,  NCoords,  transformMat, standardProt);
  computeRelCoords (relCpCoords, CpCoords, transformMat, standardProt);
  //  computeRelCoords (relHCoords,  HCoords,  transformMat, standardProt);
  computeRelCoords (relOCoords,  OCoords,  transformMat, standardProt);
  computeRelCoords (relCentroidCoords, centroidCoords, 
		    transformMat, standardProt);
}

/*
  reads filename and stores the coordinates in the relevant vectors
  coords are in homogenous coords.
*/

void readParsedPdbFile (const char * filename,
			vector<vector<float> > & NCoords, 
			vector<vector<float> > & CACoords, 
			vector<vector<float> > & CpCoords, 
			vector<vector<float> > & HCoords, 
			vector<vector<float> > & OCoords, 
			vector<vector<float> > & centroidCoords)
{
  //  FILE * fin = fopen (filename, "r");
  char          atomName[100], AAName[100];
  float         x,y,z;
  vector<float> oneCoord;
  int AAnumber;
  ifstream fin(filename);


  if (fin.good())
    {
      cout << "reading " << filename << endl;
    }
  else
    {
      cout << "could not read " << filename << endl;
      exit(0);
    }

  NCoords.resize(0);  HCoords.resize(0);
  CACoords.resize(0); CpCoords.resize(0);
  OCoords.resize(0);  centroidCoords.resize(0);
  
  oneCoord.resize(4);
  oneCoord[0] = 1;
  while (!fin.eof())
    {
      fin >> atomName  >> AAName  >> AAnumber >> x >>  y >> z;
      oneCoord[1] = x; oneCoord[2] = y; oneCoord[3] = z;

      if (!fin.eof())
	{
	  //	  cout << atomName  << " " << AAName  <<" " << x << " " << y <<" " << z << endl;
	  switch (atomName[0])
	    {
	    case 'H' :
	      {
		HCoords.push_back(oneCoord);
		break;
	      }
	    case 'N' : 
	      {
		NCoords.push_back(oneCoord);
		break;
	      }
	    case 'C' :
	      {
		// differentiating between CA and C(prime)
		if (strlen(atomName) == 2)
		  {
		    assert (atomName[1] == 'A');
		    CACoords.push_back(oneCoord);
		  }
		else
		  {
		    assert (strlen(atomName) == 1);
		    CpCoords.push_back(oneCoord);
		  }
		break;
	      }
	    case 'O' : 
	      {
		OCoords.push_back(oneCoord);
		break;
	      }
	    case 'c' :
	      {
		assert (!strcmp("centroid", atomName));
		centroidCoords.push_back(oneCoord);
		break;
	      }
	    default : 
	      {
		cout << "could not recognize atom name " << atomName << endl;
	      }
	    }
	  //	  cout << atomName << " " << AAName << " " << x << " " << y;
	  //	  cout << z << endl;
	}
    }

  cout << "num O "  << OCoords.size()  << endl;
  cout << "num H "  << HCoords.size()  << endl;
  cout << "num CA " << CACoords.size() << endl;
  cout << "num Cp " << CpCoords.size()  << endl;
  cout << "num N  " << NCoords.size()  << endl;
  
}



/* 
   reads parsed pdb file, and calls routines that
     compute relative N, CA, C', O, centroid,H coords.
   calls routines that write these coords to files in the
   current dir.

   the relative atom coords are computed wrt the vectors
       of the standard protein. 

   note: these rel. coords are different from the rel. coords 
         that would be computed for the original protein.
*/

void computeRelativeAtomCoords (char * path2DataFiles, char * proteinName) 
{
  char      parsedDsspFileName[100];
  Protein * myProtein, * standardProt;
  mat4      transformMat;
  char      path2DataFilesPlusProteinName[100];

  vector<vector<float> >  NCoords, CACoords, CpCoords,
    HCoords, OCoords, centroidCoords;
  
  vector<vector<float> >  relNCoords, 
    relCACoords, relCpCoords, 
    relHCoords, relOCoords, 
    relCentroidCoords;
 
  strcpy (path2DataFilesPlusProteinName, path2DataFiles);
  strcat (path2DataFilesPlusProteinName, proteinName);

  strcpy (parsedDsspFileName, path2DataFilesPlusProteinName);
  strcat (parsedDsspFileName, ".parsedDSSP");

  myProtein   =  new Protein(0);
  myProtein  ->  readParsedDSSPFile (&parsedDsspFileName[0]);


  char    parsedPDBFile[100];
  strcpy (parsedPDBFile, path2DataFiles);
  strcat (parsedPDBFile, proteinName);
  strcat (parsedPDBFile, ".parsedPDB");

  readParsedPdbFile  (parsedPDBFile, NCoords, CACoords, CpCoords, 
		      HCoords, OCoords, centroidCoords);
 

 
  myProtein  ->  computeSSEVectors  (CACoords);
  
  standardProt   = new  Protein(1);
  
  standardizeNativeStructure        (standardProt, myProtein);

  // don't generate CAlpha coords of the original protein
  // before copying it to standard protein since copyconfig
  // also copies the CA coordinates and then generateCACoordinates
  // called on standardProtein is returned w/o generating any 
  // CA coordinate.
  
  computeTransformationMatrix      (transformMat, standardProt, myProtein);

  computeRelCoords4AllAtoms (relNCoords, relCACoords, relCpCoords, 
			     relHCoords, relOCoords, relCentroidCoords,
			     NCoords, CACoords, CpCoords,
			     HCoords, OCoords, centroidCoords,
			     transformMat, standardProt);
  
  writeAllCoords2Files(relNCoords, relCACoords, relCpCoords,
		       relHCoords, relOCoords , relCentroidCoords);

}

/* 
   returns the SSE (0.. #SSEs - 1) corresponding to AAindex ( = 1.. numAAs in protein). 
   If not found, returns -1. 
*/
int findSseElementIndex(const Protein * protein, int AAindex) 
{
  
  bool found = false;
  int numAASoFar = Protein::firstAAIndexOfFirstNonLoopVector,
    currentSseElementIndex = 0; // 0.. # sse Elements - 1
  
  //  cout << "AAindex = " << AAindex << " numAAs = " << protein->getNumAAs() << endl;
  assert((AAindex > 0) && (AAindex <= protein->getNumAAs()));
  
  if ((AAindex <= Protein::firstAAIndexOfFirstNonLoopVector) || 
      (AAindex >  Protein::lastAAIndexOfLastNonLoopVector_plusOne)) {

    return SSE_PART_OF_REMOVED_LOOP;
  }
  else {
    while ((AAindex > numAASoFar) && (currentSseElementIndex < protein->getNumSseElements())) {
      
      if (AAindex <= (numAASoFar + ((protein->getSseElement(currentSseElementIndex))->getNumAAs()))) {
	  // found the corresponding sse Element
	  found = true;
	  break;
      }
      else {
	  numAASoFar +=  (protein->getSseElement(currentSseElementIndex))->getNumAAs();
	  currentSseElementIndex ++;
      }
    }
    if (found)
      return currentSseElementIndex;
    else return -1;
  }
}
/* computes the Xformation from a coordinate system to another */

void computeTransformationMatrix (mat4 transformMat, 
				  const Protein * stanProtein, 
				  const Protein * myProtein) 
{
  Point point1, point2, point3, point4;
  mat4 mat_s, mat_o, inv_mat_o;
  float * ptr2InvMatrixO, *ptr2MatrixO;
  
  ptr2InvMatrixO = &inv_mat_o[0][0];
  ptr2MatrixO    = &mat_o    [0][0];

  //int i,j; 
  
  assert(myProtein->getNumSseElements() >=3);
  VecCopy(point1, myProtein->getSseElement(0)->getBasePoint());
  VecCopy(point2, myProtein->getSseElement(0)->getEndPoint());
  VecCopy(point3, myProtein->getSseElement(1)->getEndPoint());
  VecCopy(point4, myProtein->getSseElement(2)->getEndPoint());

  mat_o[0][0] = 1.0; mat_o[0][1] = 1.0; mat_o[0][2] = 1.0; mat_o[0][3] = 1.0; 
  mat_o[1][0] = point1[0];   mat_o[2][0] = point1[1];   mat_o[3][0] = point1[2];
  mat_o[1][1] = point2[0];   mat_o[2][1] = point2[1];   mat_o[3][1] = point2[2];
  mat_o[1][2] = point3[0];   mat_o[2][2] = point3[1];   mat_o[3][2] = point3[2];
  mat_o[1][3] = point4[0];   mat_o[2][3] = point4[1];   mat_o[3][3] = point4[2];


  matrixInverse(ptr2InvMatrixO, ptr2MatrixO, 4);

  VecCopy(point1, stanProtein->getSseElement(0)->getBasePoint());
  VecCopy(point2, stanProtein->getSseElement(0)->getEndPoint());
  VecCopy(point3, stanProtein->getSseElement(1)->getEndPoint());
  VecCopy(point4, stanProtein->getSseElement(2)->getEndPoint());

  mat_s[0][0] = 1.0; mat_s[0][1] = 1.0; mat_s[0][2] = 1.0; mat_s[0][3] = 1.0; 
  mat_s[1][0] = point1[0];   mat_s[2][0] = point1[1];   mat_s[3][0] = point1[2];
  mat_s[1][1] = point2[0];   mat_s[2][1] = point2[1];   mat_s[3][1] = point2[2];
  mat_s[1][2] = point3[0];   mat_s[2][2] = point3[1];   mat_s[3][2] = point3[2];
  mat_s[1][3] = point4[0];   mat_s[2][3] = point4[1];   mat_s[3][3] = point4[2];


  matMult(transformMat, mat_s, inv_mat_o);
  
  cout << "transformation matrix computed. it is : " ;
  for (int i = 0; i < 4; i++) 
    {
      for (int j = 0; j < 4; j++) 
	cout << transformMat[i][j] << " ";
      cout << endl;
  }
}

#if 0
void printPairwiseCAlphaDistances(Protein * myProtein)
{
  int i,j,numCAlphas;
  
  numCAlphas   = Protein::lastAAIndexOfLastNonLoopVector_plusOne -  
    Protein::firstAAIndexOfFirstNonLoopVector;

   
  for (i = 0; i < numCAlphas; i++) 
    {
      for (j = i+1; j < numCAlphas; j++) 
	{
	  cout << "cAlphas : " << i << " and " << j ;
	  cout << " at : " << endl; 
	  printPoint(myProtein->cAlphaCoords[i]);
	  printPoint(myProtein->cAlphaCoords[j]);
	  cout << "distance : " << 
	  sqrt(myProtein->distanceSquaredBetweenCAlphas(i,j)) << endl;
	}
    }
}
#endif

/* 
   output files:    cAlphaRelDict.txt,centroidRelDict.txt,
                    CprimeRelDict.txt,NRelDict.txt,ORelDict.txt
*/

int main(int argc, char * argv[]) 
{
  if (argc < 3) 
    {
      cout << "usage : " << argv[0] << 
	" <path to pdb,dssp files> <protein_name>" << endl;
      exit(0);
    }


  char path2DataFiles[100], proteinName[100];

  strcpy ( path2DataFiles, argv[1]);
  strcpy ( proteinName,    argv[2]);


  // appending '/' to the "path to pdb, dssp files", if necessary.
  int len = strlen (path2DataFiles);
  
  if (path2DataFiles[len - 1] != '/')
    {
      cout << "appending '/' to the first parameter " << endl;
      strcat (path2DataFiles, "/");
    }

  computeRelativeAtomCoords (path2DataFiles, proteinName);
  
}
