#!/usr/bin/perl -w

# given an input (PDB) file, parses it to read the atom coords.
# this will be used to read the relative atom coords that will
# be used in generating arbitrary configs.
# only writes N, CA, C', O, HA and centroid coordinates.
# I assume HA is the H attached to CA. 
# centroid coordinates are stored in an array.


$inputFile         = shift;
unless ($inputFile =~ /\.pdb$/)
  {
    die("error! file should be a pdb file and should finish with .pdb\n");
  }

open (FIN, $inputFile) || 
    die ("couldn't open $inputFile (should be the pdb file)");


$centroid[0] = 0.0; $centroid[1] = 0.0; $centroid[2] = 0.0;
#die( "set prevAAnumber to AA number of the first AA. exiting..\n");
print STDERR  "set prevAAnumber to AA number of the first AA. \n";
$prevAANumber   = 1; # to determine when to start computing a new 
print STDERR  "currently set to $prevAANumber \n";

#centroid coordinate.
#41 for 1VII
$AAweight       = 0.0; 
$firstLinePrinted = 0;
#x,y and z coords of the centroid.
while ($line = <FIN>)
{

  if ($line =~ /^ATOM/)
    {

      &parseLine ($line);
    }
}
&writeCentroidAndUpdateParams();

# the line contains the coordinate info, which will be printed.
sub parseLine 
{
    my $line           = shift;


    unless ($line =~ /ATOM/)
    {
	die ("line = $line does not start with ATOM!\n");
    }
    
    if ($line =~ /^ATOM\s+\d+\s+(\S+)\s+(\S+)\s+[\S\s]\s+(\-*\d+)\s+(\-*\d+\.\d+)\s+(\-*\d+\.\d+)\s+(\-*\d+\.\d+)/)
    {

#	print STDERR "READ LINE = $line\n";
	$atomCode              = $1;
	$aaCode                = $2;
	$aaNumber              = $3;
	$x                     = $4;
	$y                     = $5;
	$z                     = $6;


	if ($firstLinePrinted == 0)
	  {
	    print STDERR "first AA # = ".$aaNumber."\n";
	    unless ($prevAANumber == $aaNumber)
	      {
		die("prev AA number != first aa number. change code.\n");
	      }
	    $firstLinePrinted = 1;
	  }
	
	if ($aaNumber == $prevAANumber)
	{
	  &updateCentroid($atomCode, $x, $y, $z, $AAweight);
	}
	else
	  {	    
	    &writeCentroidAndUpdateParams();
	  }
	$prevAANumber = $aaNumber;
	
	if (&writeAtom($atomCode))
	  {
	    
	    #	  print STDERR "atom code = $atomCode, aa = $aaCode, x = $x, y = $y, z = $z\n";
	    #	  if ($atomCode eq "CA")
	    #	    {
	    print STDOUT "$atomCode $aaCode $aaNumber $x $y $z\n";
	    #	  print STDOUT "$x $y $z\n";
	    #	    }
	  }
      }
    else
      {
	print STDERR "line = $line does not match the format.\n";
      }
  }

# if atom code is not of the ones to be outputted, returns 0;
# otherwise, returns the atom code.
# the recognized atoms are N, CA, C, O and HA.

sub writeAtom
{
    my $atomCode = shift;
    my $retval   = 1;
    
#    if ($atomCode =~ /^\d*(\D)/)
#    {
#	$oneLetterAtomCode = $1;
#    }
#    else
#    {
#	print "one letter atom code couldn't be extracted from $atomCode\n";
#    }
    
    unless (($atomCode eq "CA") || ($atomCode eq "N") || 
	    ($atomCode eq "O")  || ($atomCode eq "C") ||
	    ($atomCode eq "HA"))
    {
	$retval = 0;
    }
    
    $retval;
}

#updates $centroid[0..2], $AAweight.
#does not consider N, C,O or HA coords in the centroid of the sidechain.
#considers CA, since it may be the case that the AA is glycine, and there is no
#other sidechain atom involved.

sub updateCentroid
{
  my ($atomCode, $x, $y, $z) = @_;
  my $atomWeight;
  
  unless (($atomCode eq "N") || 
	  ($atomCode eq "O")  || ($atomCode eq "C") ||
	  ($atomCode eq "HA"))
    {
	
      if ($atomCode =~ /^N/)
	{
	  $atomWeight  = 14;
	}
      elsif ($atomCode =~ /^C/)
	{
	  $atomWeight  = 12;
	}
      elsif ($atomCode =~ /^O/)
	{
	  $atomWeight  = 16;
	}
      elsif (($atomCode =~ /^\wH/) || ($atomCode =~ /^H/))
	{
	  $atomWeight  = 1;
	}
      elsif ($atomCode =~ /^S/) 
	{
	  $atomWeight  = 32;
	}
      else
	{
	  print STDERR "unknown atom code = ".$atomCode."\n";
	}
      $AAweight    += $atomWeight;
      $centroid[0] += $x * $atomWeight;
      $centroid[1] += $y * $atomWeight;
      $centroid[2] += $z * $atomWeight
	
    }
}

# writes the centroid, and then resets centroid and aaweight parameters
sub writeCentroidAndUpdateParams()
{
    my $i;

    

    for ($i = 0; $i < 3; $i++)
    {
	$centroid[$i] = $centroid[$i] / $AAweight;
    }
    print STDOUT "centroid _  $prevAANumber $centroid[0] $centroid[1] $centroid[2]\n";
    $AAweight     = 0;
    $centroid[0]  = $centroid[1] = $centroid[2] = 0.0;
    &updateCentroid($atomCode, $x, $y, $z);
}
