#ifndef _SSE_H
#define _SSE_H
#ifndef _MYVECTOR_H
  #include "myVector.H"
#endif
//#include <stream.h>
#ifndef  _PROTEIN_H
  #include "protein.H"
#endif
#include <vector>

enum sseType {ALPHA,BETA,LOOP};

class Protein;
class Sse {
 public:
  
  Sse (const sseType itype,const Point ibase, const Point iend, int inumAAs);

  Sse (const sseType itype, const vector<float> & ibase, 
       const vector<float> & iend,  int inumAAs) ;
  
  /* the function below compute the tip of the vector given the parameters. Also sets theta, phi. */
  void computeVector (const Point& ibase, float ilength, float vectorAngle, float dihedralAngle, 
		      Sse * prevSse, Sse * prevprevSse);

  /* a second method for computing vector. Calls the above method. */
  void computeVector (int sseIndex, float vectorAng, float dihedralAng,
		      float twistAngle, float length, Protein * fatherProt);

  /* friend declarations */
  friend ostream& operator<< (ostream & output, Sse* s);
  friend void     Protein::computeLoopVectors();

  /* mutators */
  void setVector2(const Point base, const Point end); /* also updates vector length.*/
  void incrementNumAAs(int);
  void setTwistAngle(float);
  void setVectorAngle(float);
  void setDihedralAngle(float);

  /* accessors */
  sseType getType() const;
  const   Point& getBasePoint() const;
  const   Point& getEndPoint() const;

  float  getTwistAngle() const;
  float  getVectorAngle() const;
  float  getDihedralAngle() const;
  float  getLength()   const;
  int    getNumAAs() const;
  
  private :
    
  sseType type;
  myVector  vec;   //     length of this vector is the hidden third DOF for loops.
  float  vectorAngle, dihedralAngle, twistAngle;
  // NOTE: these facts are used in normalize function
  // that normalizes these parameters between 0 and 1.
  // vectorAngle     : 0.. pi
  // dihedralAngle : -pi .. pi 
  // twistAngle    : -pi <= twistAngle <= pi

  /* suppose the base point of this SSE is P1, end point is P2. Suppose the previous SSE base point is P0. Its end pt. is P1. Suppose the SSE 2 before the current SSE has base P(-1), end pt. P0. vectorAngle is the angle between P0, P1 and P2. Dihedral_angle is the angle between P(-1), P0, P1 and P2.
   */
  int     numAAs;
  /*
    phi   : -pi < phi <= pi
    theta : 0 .. pi
  */
  
};

#endif
