#include "vec_utility.H"
#include <iostream>
//#include "common.h"
#include <assert.h>
#include <stdio.h>

void printPoint(Point A) 
{
  for (int i = 0;  i < 3; i++) 
    {
      cout << A[i] << " ";
    }
  cout << endl;
}

/* computes C = a A + B */
void VecAddS(Vec C, float a, const Vec A, const Vec B) {

  C[0] = a * A[0] + B[0];
  C[1] = a * A[1] + B[1];
  C[2] = a * A[2] + B[2];

}

/* computes Result = a * A */
void VecScale(Vec Result,const float a, const Vec A) {
  int i;


  /*  printf("vec scale: %f\n",a);*/

  for (i=0; i<3; i++)
    Result[i] = a * A[i];
}


void VecAdd(Vec Result,const Vec A, const Vec B) {
  int i;
  for (i=0; i<3; i++)
    Result[i] = A[i] + B[i];
}


void VecCombine(Vec Result,float a, const Vec A, float b, const Vec B) {

  Vec  Va,Vb;

  /*  printf("vec combine: %f %f\n",a,b);*/

  VecScale(Va,a,A);
  VecScale(Vb,b,B);
  VecAdd(Result,Va,Vb);
}

/* computes A-B = Result */
void VecSub(Vec Result, const Vec A, const Vec B) {

  int i;
  for (i=0; i<3; i++)
    Result[i] = A[i] - B[i];
}

float VecLength(const Vec P)
{
  float length = 0.;
  int i;
  
  for (i=0; i<3; i++)
    length += P[i] * P[i];

  length = sqrt(length);
  return length;
}

/* dest = source */
void VecCopy(Vec dest, const Vec source) {

     dest[0] = source[0];
     dest[1] = source[1];
     dest[2] = source[2];
}

/* at the end, either P has unit length or 0 length */
void VecNormalize(Vec P)
{
  float length = 0.;
  int i;
  
  length = VecLength(P);

  if (length != 0.0) {
    for (i=0; i<3; i++)
      P[i] /= length;
  }
  else {
    cout << "P had 0 length so couldn't normalize it. " << endl;
  }
}

float VecDot (const Vec A, const Vec B) {
  int i;
  float retVal=0.;

  for (i=0; i<3; i++)
    retVal += A[i] * B[i];
  return retVal;
}



/* component-wise multiplication */
void VecMul(Vec Result,const Vec A, const Vec B) 
{
  int i;
  for (i=0; i<3; i++) {
    Result[i] = A[i] * B[i];
  }
}
/* return the cross product Result = A cross B */
void VecCross(Vec Result,const Vec A, const Vec B) {

	Result[0] = (A[1]*B[2]) - (A[2]*B[1]);
	Result[1] = (A[2]*B[0]) - (A[0]*B[2]);
	Result[2] = (A[0]*B[1]) - (A[1]*B[0]);
}

int VecEqual (const Vec A, const Vec B) {

  if (A[0] != B[0]) return 0;
  if (A[1] != B[1]) return 0;
  if (A[2] != B[2]) return 0;
  return 1;
}



/*
      P1[0] P2[0] P3[0]
det ( P1[1] P2[1] P3[1] ) is returned.
      P1[2] P2[2] P3[2]

*/

float determinant(const Vec P1, const Vec P2, const Vec P3) {
  
  float retVal;

  retVal  = P1[0] * (P2[1] * P3[2]- P3[1] * P2[2]);
  retVal -= P2[0] * (P1[1] * P3[2]- P1[2] * P3[1]);
  retVal += P3[0] * (P1[1] * P2[2] -P2[1] * P1[2]);

  return retVal;
}


void  obtainVecFromPoints(const Point& p_base, const Point& p_end, Vec& v) {
  v[0] = p_end[0] - p_base[0];
  v[1] = p_end[1] - p_base[1];
  v[2] = p_end[2] - p_base[2];
}


// angles in radian, length and coordinates in angstrom
/* first handles dihedral Angle, then bond Angle */
void  obtainFourthPoint(const Point& pA, const Point& pB, const Point& pC, Point  pD,  float dihedral_angle, const float bond_angle, const float bond_length) {


  vec4 A,C,D; //B
  float alpha, cos_alpha;
  const float epsilon = 0.03;
  mat4 B_to_O = { {1,0,0,0}, {-pB[0],1,0,0}, {-pB[1],0,1,0} ,{-pB[2],0,0,1}};
  mat4 O_to_C = { {1,0,0,0}, {pC[0],1,0,0} , {pC[1],0,1,0}  ,{pC[2],0,0,1 }};

  if (dihedral_angle < 0 ) {
    dihedral_angle += 2 * M_PI;
  }
  float a = cos(dihedral_angle/2);
  for (int i = 1; i< 4; i++) {
    A[i] = pA[i-1];
    //    B[i] = pB[i-1];
    C[i] = pC[i-1];
  }
  A[0] = /*B[0]=*/ C[0] = 1;

  // A,B,C vec4's are ready now.
  
  float b = pC[0] - pB[0];
  float c = pC[1] - pB[1];
  float d = pC[2] - pB[2];
  //  float denum = 1.0;
  float coeff;
  assert(!VecEqual(pB,pC)); // making sure b*b + c*c+ d*d != 0.0
  coeff = sqrt((1 - a*a) / (b*b + c*c+ d*d));
  b = b * coeff; c = c * coeff; d = d * coeff;

  assert(fabs(a*a + b*b + c*c + d*d - 1.0) < epsilon); // remove this later
  mat4 Q1 = {{1.0,  0., 0., 0.}, 
	     {0, a*a + b*b - c*c - d*d, 2*b*c - 2*a * d, 2*b*d + 2*a*c}, 
	     {0, 2 * (b * c + a * d), a*a - b*b + c*c - d*d, 2 * ( c * d - a * b)}, 
	     {0, 2 * (b * d - a * c), 2 * (c * d + a * b), a*a - b*b - c*c + d*d}};
  
  mat4 tempMat4, transformation;
  matMult(tempMat4,Q1, B_to_O);
  matMult(transformation,O_to_C, tempMat4);
  applyMatToVec4(D,    transformation, A);

  vec4 unitV;
  vec4Sub(unitV, D,C);
  vec4Normalize(unitV);
  vec4Combine(D, bond_length, unitV, C);


  Vec CB,CD;
  VecSub(CB, pB, pC);
  pD[0] = D[1]; pD[1] = D[2];pD[2] = D[3];
  VecSub(CD, pD, pC);

  float VecLength_CB = VecLength(CB),
    VecLength_CD = VecLength(CD);

  /*  
      assert(VecLength_CB != 0.0); this is already checked in the above assert 
      (in making sure b*b +c*c + d*d != 0.0) 
  */
  assert(VecLength_CD != 0.0);
  cos_alpha = VecDot(CB, CD) / VecLength_CB  / VecLength_CD ;

  if (cos_alpha > 1.0) {
    assert(cos_alpha < 1.0 + epsilon);
    alpha = 0.0;
  }
  else if (cos_alpha < -1.0) {
    assert(cos_alpha >= -1.0-epsilon);
    alpha = M_PI;
  }
  else 
    alpha = acos  (cos_alpha);

  float delta_alpha = bond_angle - alpha;
  while (delta_alpha < 0) {
    delta_alpha += 2 * M_PI;
  }

  Vec  vector_of_revolution ;
  VecCross(vector_of_revolution,CB,CD);

  a = cos(delta_alpha /2);
  b = vector_of_revolution[0];c = vector_of_revolution[1];d = vector_of_revolution[2];

  if (b * b + c * c + d * d == 0.0) { // if BC and CD are aligned
    //       select vec_of_revolution as an arbitrary vector in the subspace -| to BC and CD. 

    if ((CD[1] != 0.0) && (CD[0] != 0.0)) {
      vector_of_revolution[0] = - CD[1]; vector_of_revolution[1] =   CD[0]; vector_of_revolution[2] = 0.0;
    }
    else {
      b = vector_of_revolution[0] = 1.0; c = vector_of_revolution[1] = d = vector_of_revolution[2] = 0.0;
    }
  }

  assert(fabs(VecDot(vector_of_revolution, CB)) < epsilon); // remove these later
  if (fabs(VecDot(vector_of_revolution, CD)) >= epsilon) {
    cout << "CD : " ;
    printPoint(CD);
    cout << " vec. of rev. : " ;
    printPoint(vector_of_revolution);
    cout << "CB : " ;
    printPoint(CB);
    cout << "alpha = " << alpha << "delta_alpha = " << delta_alpha << endl;
    cout << " vec. dot = " << fabs(VecDot(vector_of_revolution, CD)) << endl;
  }

  assert(fabs(VecDot(vector_of_revolution, CD)) < epsilon);

  coeff = sqrt((1 - a*a) / (b*b + c*c+ d*d));
  
  b = b * coeff; c = c * coeff; d = d * coeff;
  
  assert(fabs(a*a + b*b + c*c + d*d - 1.0) < epsilon); // remove this later
  mat4 Q2 = {{1.0, 0, 0, 0}, 
	     {0, a*a + b*b - c*c - d*d, 2*b*c - 2*a * d, 2*b*d + 2*a*c}, 
	     {0, 2 * (b * c + a * d), a*a - b*b + c*c - d*d, 2 * ( c * d - a * b)}, 
	     {0, 2 * (b * d - a * c), 2 * (c * d + a * b), a*a - b*b - c*c + d*d}};
 
  mat4 C_to_O = { {1,0,0,0}, {-pC[0],1,0,0} , {-pC[1],0,1,0}  ,{-pC[2],0,0,1 }};
  

  matMult(tempMat4,Q2 , C_to_O);
  matMult(transformation,O_to_C, tempMat4);
  vec4 tempD;
  applyMatToVec4(tempD, transformation, D);
  
  pD [0] = tempD[1];
  pD [1] = tempD[2];
  pD [2] = tempD[3];
  
}

void matMult(mat4 myoutmat, const mat4 mymat1, const mat4 mymat2) {

  int i, j;

  for (i = 0; i < 4; i++) {
    for (j = 0; j < 4; j++) {    
      myoutmat[i][j] = mymat1[i][0]*mymat2[0][j] + mymat1[i][1]*mymat2[1][j] + mymat1[i][2]*mymat2[2][j] + mymat1[i][3]*mymat2[3][j];
    }
  }
}
/*
void mat3Mul(mat3 myoutmat, const mat3 mymat1, const mat3 mymat2) {

  int i, j;

  for (i = 0; i < 3; i++) {
    for (j = 0; j < 3; j++) {
      myoutmat[i][j] = mymat1[i][0]*mymat2[0][j] + mymat1[i][1]*mymat2[1][j] + mymat1[i][2]*mymat2[2][j];
    }
  }
}
*/
void applyMatToVec4(vec4 outvec,  const mat4 themat, const vec4 invec) 
{

  outvec[0] = invec[0]*themat[0][0] + invec[1]*themat[0][1] + invec[2]*themat[0][2] + invec[3] * themat[0][3];
  outvec[1] = invec[0]*themat[1][0] + invec[1]*themat[1][1] + invec[2]*themat[1][2] + invec[3] * themat[1][3];
  outvec[2] = invec[0]*themat[2][0] + invec[1]*themat[2][1] + invec[2]*themat[2][2] + invec[3] * themat[2][3];
  outvec[3] = invec[0]*themat[3][0] + invec[1]*themat[3][1] + invec[2]*themat[3][2] + invec[3] * themat[3][3];
}

void applyMatToVec4 (vector<float> &        outvec, 
		     const             mat4 themat, 
		     const vector<float> &  invec) 
{

  outvec[0] = invec[0]*themat[0][0] + invec[1]*themat[0][1] + 
    invec[2] * themat[0][2] + invec[3] * themat[0][3];

  outvec[1] = invec[0]*themat[1][0] + invec[1]*themat[1][1] + 
    invec[2] * themat[1][2] + invec[3] * themat[1][3];

  outvec[2] = invec[0]*themat[2][0] + invec[1]*themat[2][1] + 
    invec[2] * themat[2][2] + invec[3] * themat[2][3];

  outvec[3] = invec[0]*themat[3][0] + invec[1]*themat[3][1] + 
    invec[2] * themat[3][2] + invec[3] * themat[3][3];
}


/*
void applyMatToVec3(vec3 outvec,  const mat3 themat, const vec3 invec) {
    
    outvec[0] = invec[0]*themat[0][0] + invec[1]*themat[0][1] + invec[2]*themat[0][2];
    outvec[1] = invec[0]*themat[1][0] + invec[1]*themat[1][1] + invec[2]*themat[1][2] ;
    outvec[2] = invec[0]*themat[2][0] + invec[1]*themat[2][1] + invec[2]*themat[2][2] ;
  
}
*/

void  vec4Combine(vec4 D, float a, const vec4 unitV, const vec4 C) {
  vec4 tempV;

  vec4Scale(tempV, a , unitV);
  vec4Add  (D, tempV, C);
}

void vec4Scale(vec4 Result, float a, const vec4 V) {
  
  for (int i = 0; i<4 ; i++){
    Result[i] = a * V[i];
  }
}

void vec4Add (vec4 Result, const vec4 A, const vec4 B) {
  for (int i = 0; i< 4; i++) {
    Result[i] = A [i] + B[i];
  }
}

void vec4Sub( vec4 Result, const vec4 V1, const vec4 V2) {

  Result[0] = V1[0] - V2[0];
  Result[1] = V1[1] - V2[1];
  Result[2] = V1[2] - V2[2];
  Result[3] = V1[3] - V2[3];
}

void vec4Normalize(vec4 V) {

  float length=vec4Length(V);
  if (length  == 0.0) {
    cout << "Vector length is 0. Can't normalize. " << endl;
  }
  else {
    V[0] /= length;
    V[1] /= length;
    V[2] /= length;
    V[3] /= length;
  }
}

float vec4Length(const vec4 v) {
  
  return (sqrt(v[0] * v[0] + v[1] * v[1] +v[2] * v[2] +v[3] * v[3] ));
}

void vec4Copy( vec4 dest, const vec4 source) {

  dest[0] = source[0];
  dest[1] = source[1];
  dest[2] = source[2];
  dest[3] = source[3];
}
  
float findRMSD(const Point p1[], const Point p2[], const int numAtoms) {

  int i;
  float retval = 0.0;
  
  for (i = 0; i < numAtoms; i++) {
    retval = retval + (p1[i][0] - p2[i][0])*(p1[i][0] - p2[i][0]) + 
             (p1[i][1] - p2[i][1])*(p1[i][1] - p2[i][1]) + 
             (p1[i][2] - p2[i][2])*(p1[i][2] - p2[i][2]);
  }

  retval = sqrt(retval/((float)(numAtoms)));

  return retval;
}

int positive_volume(Point A,Point B, Point C, Point D) {
// input A,B,C,D = four points 
// output = 1 (TRUE) if sign of dihedral angle should be positive, else 0 (FALSE) 
// algorithm compares relative orientation of ABC plane to D vector to 
// determine if D vector is in positive half of plane or negative half. 

  Point p21,p23,cp1;
  double test1,test2;
  Point pTest1, pTest2, temp;


  VecSub(p21,A,B);
  VecSub(p23,C,B);
  VecCross(cp1,p23,p21);
  VecSub(temp,B,cp1);
  VecSub(pTest1,D,temp);
  test1 = VecLength(pTest1);

  VecAdd(temp,B,cp1);
  VecSub(pTest2,D,temp);
  test2 = VecLength(pTest2);
  
  if (test1 > test2) 
    return 1;
  else return 0;
}

double dihedral_angle (Point A, Point B, Point C, Point D) {

  //input: A, B, C, D are (x y z) points 
  //output: dihedral angle. Looking down BC axis, angle between AB and CD. 
  // ranges from -PI to PI. AB axis is at 0, with positive angles 
  // clockwise, and negative angles counterclockwise. 
  // 
  // "x" operator, as in "d12 x d23" is cross product of vector d12 and vector d23 
  // "*" operator is the dot-product 
  // 

  //  const double _180_div_PI = 57.29578;
  Point d12, d23, d34, x123, x234;
  double cosphi,denum;
  double length_x123_sqr, length_x234_sqr;
  const float epsilon = 1E-4;

  VecSub(d12,A,B);
  VecSub(d23,B,C);
  VecSub(d34,C,D);
  VecCross(x123,d12,d23);
  VecCross(x234,d23,d34);

  length_x123_sqr  = x123[0] * x123[0] + x123[1] * x123[1] + x123[2] * x123[2];
  if (length_x123_sqr < epsilon)
    return 0.0;
  

  length_x234_sqr  = x234[0] * x234[0] + x234[1] * x234[1] + x234[2] * x234[2];
  if (length_x234_sqr < epsilon)
    return 0.0;

  denum = sqrt(length_x123_sqr * length_x234_sqr);

  assert(denum != 0.0);
  cosphi   = VecDot(x123,x234) / denum;
  if (positive_volume(A,B,C,D)) {
    return acos(cosphi);
  }
  else {
    return -acos(cosphi);
  }
}

// Function written by Hesham Eldeeb, 1999. Taken from www.eldeeb.com
//////////////////////////////////////////////////////////////////////////
//
// pMatrix: pointer to array of doubles containing the matrix elements
//          arranged row-wise
//
// pInverse: pointer to array of doubles that will contain the inverse
//           elements arranged row-wise
//
// iRank: rank of matrix
//
//////////////////////////////////////////////////////////////////////////
// returns 0 if successful, -1 o.w. 
float matrixInverse( float * pInverse, float * pMatrix, int iRank )
{
 const int MAX_RANK = 7;
 int i, j, k ;
 int bOk ;
 float dTemp, dCoeff ;
 float MatCopy[ MAX_RANK * MAX_RANK ] ;
 float * pMatCopy ;



 //-------------

 pMatCopy = &MatCopy[ 0 ] ;

 for( i = 0 ; i < iRank ; i ++ )
  for( j = 0 ; j < iRank ; j ++ )
   pMatCopy[ i * iRank + j ] = pMatrix[ i * iRank + j ] ;

 //-------------

 for( i = 0 ; i < iRank ; i ++ )
  for( j = 0 ; j < iRank ; j ++ )
   pInverse[ i * iRank + j ] = 0.0 ;

 for( i = 0 ; i < iRank ; i ++ )
  pInverse[ i * iRank + i ] = 1.0 ;

 //-------------


 // main loop
 for( i = 0 ; i < iRank ; i ++ )
 {
  // make sure pivot is non-zero
  if( ( fabs( pMatCopy[ + i * iRank + i ] ) < 0.0001 ) )
  {
   bOk = 0 ;
   for( j = i + 1 ; j < iRank ; j ++ )
   {
    if( fabs( pMatCopy[ + i * iRank + j ] ) >= 0.0001 )
    {
     // swap columns i & j in matrix
     for( k = 0 ; k < iRank ; k ++ )
     {
      dTemp = pMatCopy[ k * iRank + j ] ;
      pMatCopy[ k * iRank + j ] = pMatCopy[ k * iRank + i ] ;
      pMatCopy[ k * iRank + i ] = dTemp ;
     }

     // swap columns i & j in inverse
     for( k = 0 ; k < iRank ; k ++ )
     {
      dTemp = pInverse[ k * iRank + j ] ;
      pInverse[ k * iRank + j ] = pInverse[ k * iRank + i ] ;
      pInverse[ k * iRank + i ] = dTemp ;
     }

     bOk = 1 ;
     break ;
    }
   }

   if( !bOk )
   {
    printf( "Matrix irregular\n" ) ;
    return( -1 ) ;
   }
  }
  //-----

  for( j = 0 ; j < iRank ; j ++ )
  {
   if( j == i )
    continue ;

   dCoeff = pMatCopy[ i * iRank + j ] / pMatCopy[ i * iRank + i ] ;

   for( k = 0 ; k < iRank ; k ++ )
    pMatCopy[ k * iRank + j ] -= pMatCopy[ k * iRank + i ] * dCoeff ;

   for( k = 0 ; k < iRank ; k ++ )
    pInverse[ k * iRank + j ] -= pInverse[ k * iRank + i ] * dCoeff ;
  }

 }
 //-----

 // normalize
 for( i = 0 ; i < iRank ; i ++ )
  for( k = 0 ; k < iRank ; k ++ )
   pInverse[ k * iRank + i ] /= pMatCopy[ i * iRank + i ] ;

 return( 0 ) ;
}
