// 3/23 note: maxEnergy modified for 1LEU run.
#ifndef _PROTEIN_H
  #include "protein.H"
#endif
#include <stdlib.h>
#include <iostream>
#include <string.h>
#include <fstream>
#include <stdio.h>
#ifndef _SSE_H
  #include "sse.H"
#endif
#ifndef  VEC_UTILITY_H
  #include "vec_utility.H"
#endif
#include <assert.h>
//#include "common.h"
#include <vector>
#ifndef  _TREE_H
  #include "tree.H"
#endif
#include <map>
#include <ctype.h>
#include "energyFunctions.H"
#include <iostream>
//#include <ANN/ANN.h>
#include <sys/times.h>

extern          int      VERBOSE_MODE;
extern unsigned int      ANN_dim; 
//extern Protein **    allNodes;
typedef Sse   * pSse;
typedef float * pFloat;
int          Protein::numElements        =  0;
//map<char,bool>                   Protein::HP_mapping;
vector<char>                     Protein::sequence;
vector<char>                     Protein::sseTypeSeq;
vector<bool>                     Protein::charged;
vector<float>                    Protein::charge;
map<char,bool>                   Protein::hydrophobic;
map<int, bool>                   Protein::loop;
vector<int>                      Protein::nonLoopAAIndices; 
const float                      Protein::maxExEnergy =  1.0; // even native structure has some exclusion energy component
//const float                      Protein::minExEnergy =  0.0;
float                            Protein::minEnergy      = -50.0;
const float                      Protein::minLengthPerAA = 0.5;  /* used by 
								    create{"Non",""}RandomConfig. 
								    only used for loops.*/
const float                      Protein::maxLengthPerAA = 6.0;
float  **                        Protein::relcAlphaCoords  = NULL; 
/*allocated by doSetup,  never deallocated. */
float  **                        Protein::relCentroidCoords = NULL; 
/* same as above. */

float  **                        Protein::relCprimeCoords = NULL;
float  **                        Protein::relNCoords = NULL;
float  **                        Protein::relOCoords = NULL;


//float                            Protein::closeEnoughDistance = 1.0; 
/* 
   if the distance between protein structures is less than this amount,
   I consider them to be connectable by a straight link. Should be 
   less than connectionDistanceThreshold. 
   See explanation w/ it. 
*/

float                            Protein::MAX_PATH_LENGTH     = 10000.0; 
const float                      Protein::ENERGY_WEIGHT       = 1.0;

/* constants taken from ldock.H exactly. */

const float                      Protein::LENGTH_WEIGHT       = 0.0;
int Protein::NUMDOF;             /* num of DOF of the protein, 
				    is computed in Protein::doSetup by 
				    assigning 3 DOF to loops, 2 DOF to alphas,
				    betas. 
				 */

//double Protein::connectionDistanceUpperbound = 100.0; 

/* 
   should be bigger than closeEnoughDistance to be able to use
   proteinDistanceCheck for protein distance comparisons with
   closeEnoughDistance.	the max. distance for trying to connect two 
   protein structures. 
*/

const float Protein::minAngleDifference  = 15 * M_PI/180; 
/* used to resolve cases in which 
						    the two structures differ 
						 */
const float Protein::minLengthDifference = 0.5; 
/* very slightly in their parameters, 
						    and yet they have a distance 
						    bigger than closeEnoughDistance
						 */
const float Protein::angle_epsilon       = 30 * M_PI/180;   /* these values are used to generate */
const float Protein::length_epsilon      = 0.5;    /* configs around original config. */
/*
  they should be bigger than minangledifference and minlengthdifference to be 
  able to differentiate one config. from another. However, they should be
  close enough to the original config. to have low enough energy. I selected
  them so that in angles,  there is ~ 15 degrees +- original config. and 
  0.5 A +- in length around 
  original config. again.
*/
const float Protein::minSampleClusterSize = 0.7;
/* while looking for nodes to expand around, the min. distance between 2 
   "seeds" should at least be this amount. Since I do not differentiate
   configs. at a distance of closeEnoughDistance or less, this parameter 
   should be > closeEnoughDistance. Also, this parameter should be such that
   a seed should not be generatable by sampling around another seed, so it 
   also depends on the angle_epsilon and length_epsilon. Since w/ current 
   paramaters it seems that 2 configs can be
   at most at a distance of ~3.6A, , if generated with createnonrandomconfig, 
   so I set this parameter to 5 A. 
*/

//const float Protein::closeThreshold = 0.05; // two proteins are considered 
//close if their distance under the new metric  is < 0.05.

int   Protein::modelNumber = 0;
float Protein::center[3] = {0.};
float Protein::objMat[16] = {0.};
int numComputations = 0;
int
  Protein::lastAAIndexOfLastNonLoopVector_plusOne = 0, 
  Protein::firstAAIndexOfFirstNonLoopVector = 0;
vector<vector<float> > Protein::pDSqr;//pairwiseDistances;
vector<HHBondType> Protein::HHBondsList;


extern void reportTimes (struct tms start, struct tms end);
void  findBaseAndUnitVecs(float base[], float unit[], Protein *myProtein, 
			  int currentSseIndex,  int relAAIndex );



int   findIndex(const char AAchar);
//unsigned int shortcut, noshortcut;

void extStr(char* out, const char* in, int st, int end) {
 
  int outcount = 0; 
  int i;

  for (i = st - 1; i < end; i++) { 
 
    out[outcount] = in[i]; 
    outcount++; 
  } 
 
  out[outcount] = '\0'; 
 
}


/* just sets elements to NULL */
Protein::Protein(int _index) {

  //elements       = NULL;
  sseVectors.resize(0);
  cAlphaCoords   = NULL;
  centroidCoords = NULL;
  NCoords  = OCoords = CPrimeCoords =  NULL;
  myTree   = NULL;
  myNode   = NULL;
  numLinks = 0;
  index    = _index;
  
  energy       = coulombicEnergy= exclusionEnergy= HHEnergy = 0.0;
  hBondEnergy  = 0.0;
  indivChargesCoulombicContribution= alphaHelixCoulombicContribution = 0.0;
  //  rmsdFromNative = 0.0;

}

/* 
   dofParameters order should be as follows : 
   dihedralAngle, vectorAngle, twistAngle, length,  x numSSeElements
   also computes energy and rmsd from nativeprotein.
*/

Protein::Protein (float * dofParameters, int numParams, 
		  Protein * nativeProtein) 
{
  
  float vectorAng, dihedralAng, twistAng, length;
  int i;
  const Point originPoint = {0.0, 0.0, 0.0};

  assert(numParams == numElements * 4);
  assert(nativeProtein->getIndex() == 0);
  
  cAlphaCoords = centroidCoords =   NCoords  = OCoords = CPrimeCoords = NULL;
  myTree   = NULL; myNode   = NULL;
  numLinks = 0; index = Protein::uninitializedIndex;
  
  //  elements      = new pSse[numElements];
  sseVectors.resize(0);
  for (i = 0; i< numElements; i++) 
    {
      Sse * ptr2Sse = 
	new Sse (nativeProtein->sseVectors[i]->getType(), 
		 originPoint, originPoint, 
		 nativeProtein->sseVectors[i]->getNumAAs());
    
      sseVectors.push_back (ptr2Sse);
    
      dihedralAng = dofParameters[i * 4];
      vectorAng   = dofParameters[i * 4 + 1];
      twistAng    = dofParameters[i * 4 + 2];
      length      = dofParameters[i * 4 + 3];

      ptr2Sse->computeVector(i, vectorAng, dihedralAng, twistAng, 
			     length, this);
    }
  
  computeEnergy();
  findParams();
  //  rmsdFromNative = computeRMSD(nativeProtein);
}

/*
Protein& Protein::operator= (Protein * original) {
  if (original != this) {
    if (original != NULL) {
      delete original;
      for (int i = 0; i< numElements; i++) {
	elements[i] = original->elements[i];
      }
      for (i = 0; i< sequence.size(); i++) {
	VecCopy(cAlphaCoords[i], original->cAlphaCoords[i]);
	VecCopy(centroidCoords[i], original->centroidCoords[i]);
	
      }
    }
  }
}
*/

Protein::~Protein() 
{
  int i,arraySize; // common arraySize for C-alpha, centroid, N,C, O arrays

  arraySize = lastAAIndexOfLastNonLoopVector_plusOne - 
    firstAAIndexOfFirstNonLoopVector;

  for (unsigned int i = 0; i< sseVectors.size(); i++) 
    {
      if (sseVectors[i] != NULL) 
	{
	  delete sseVectors[i];
	}
    }
  
  sseVectors.resize(0);

  if (cAlphaCoords != NULL) {
    for (i = 0; i< arraySize; i++) {
      delete[] cAlphaCoords[i];
    }
    delete[] cAlphaCoords;
  }
  
  if (myNode != NULL) {
    delete myNode;
  }

  if (centroidCoords != NULL) {
    for (i = 0; i< arraySize; i++) {
      delete[] centroidCoords[i];
    }
    delete[] centroidCoords;
  }
  
  if (NCoords != NULL) {
    for (i = 0; i < arraySize; i++) {
      delete[] NCoords[i];
    }
    delete[] NCoords;
  }

  if (OCoords != NULL) {
    for (i = 0; i < arraySize; i++) {
      delete[] OCoords[i];
    }
    delete[] OCoords;
  }
  
  if (CPrimeCoords != NULL) 
    {
      for (i = 0; i < arraySize; i++) {
	delete[] CPrimeCoords[i];
      }
      delete[] CPrimeCoords;
    }
}

void extract(Point point, const char * line) 
{
  char record[10];
  
  extStr(record,line,117,122);
  point[0] = (float) (atof(record));
  extStr(record,line,124,129);
  point[1] = (float) (atof(record));
  extStr(record,line,131,136);
  point[2] = (float) (atof(record));
}

/* creates and returns a sse between p1 and p2 */
Sse * loopVector (const vector<float> & p1, const vector<float> & p2,
		  int numAAs)
{
  Sse * retval = new Sse (LOOP, p1, p2, numAAs);
  return retval;
}

/*
  computes and returns the alpha helix vector, which extends from
  AA #k to AA #l.
*/

Sse * alphaHelixVector (const vector<vector<float> > &cACoords, int k, int l)
{
  Sse * retval;
  Point vecBase, vecEnd;
  int   numAAs = l - k + 1;
  assert (numAAs >= 4);
  int i;

  if (numAAs == 4) 
    {
      assert (k > 0);
      // if this fails, implement the routine to pick the extra point from
      // the other end point of the alpha helix
      for (i = 0; i<3; i++) 
	{
	  vecBase[i] = (0.74 * cACoords[k-1][i] + cACoords[k][i]  + 
			cACoords[k+1][i] + 0.74 * cACoords[k+2][i]) / 3.48;
	  vecEnd[i]  = (0.74 * cACoords[k+3][i] + cACoords[k+2][i]  + 
			cACoords[k+1][i] + 0.74 * cACoords[k][i]) / 3.48;
	}
    }
  else
    {
      for (i=0; i<3; i++) 
	{
	  vecBase[i] = (0.74 * cACoords[k][i] + cACoords[k+1][i]  + 
			cACoords[k+2][i] + 0.74 * cACoords[k+3][i]) / 3.48;
	  vecEnd[i]  = (0.74 * cACoords[l][i] + cACoords[l-1][i]  + 
			cACoords[l-2][i] + 0.74 * cACoords[l-3][i]) / 3.48;
	}
    }
  retval          = new Sse(ALPHA, vecBase, vecEnd, numAAs);
  
  return retval;
}

/*
  computes the beta strand vector, and returns it
*/
Sse * betaStrandVector (const vector<vector<float> > &cACoords, int k, int l)
{

  int numAAs = l - k + 1;
  assert (numAAs >= 3);
  Sse * retval;
  Point vecBase, vecEnd;
  
  for (int i = 0; i < 3; i++) 
    {
      vecBase[i] = (cACoords[k][i] + cACoords[k + 1][i]) / 2.0;
      vecEnd [i] = (cACoords[l][i] + cACoords[l - 1][i]) / 2.0;
    }
  
  retval     = new Sse(BETA,vecBase,vecEnd, numAAs);
  return retval;

}

#if 0
/*  
    I will save a dummy vector for loops and then change those vector
base and end configurations in order to make this vector a "dashed"
one, connecting real secondary structure vectors.  
*/

Sse * saveAsLoop(int *lastSavedIsLoop, Point fp, Point lp, int numAtoms) 
{
  //  Point vecBase,vecEnd;
  Sse * retVal;
  //  Point point1,point2;
  //  int i;

  if ((*lastSavedIsLoop) == 1) 
    {
      // don't save again...
      // and still keep the flag as 1.
      retVal = NULL;
    }
  else 
    {
      /*
	for (i=0; i<3; i++) {
	point1[i] = fp[i];
	point2[i] = lp[i];
	}
	retVal = new Sse(LOOP, point1, point2) ;
      */
      retVal = new Sse(LOOP, fp, lp, numAtoms) ;
      (*lastSavedIsLoop) = 1;
    }
#if 0
  retVal = NULL;
  if (numAtoms == 1) {
    // selecting the aa. from last line if it exists as the base, if it does not exist, selecting the aa. from next line.
    if (strcmp(prevLine,"")) {
      cout << "prevLine = " << prevLine << endl;
      extract(vecBase,prevLine);
      retVal     = new Sse(LOOP,vecBase,points[0]);
    }
    else if (strcmp(nextLine,"")) {
      cout << "nextLine = " << nextLine << endl;
      extract(vecEnd,nextLine);
      retVal     = new Sse(LOOP,points[0],vecEnd);
    }
    else {
      cout << "This situation should not arise." << endl;
    }
  }
  else {
    // selecting the first amino acid as the base and the last one as the destination
    retVal = new Sse(LOOP,points[0],points[numAtoms-1]);
  }
#endif
  return retVal;
}
#endif

#if 0

Sse * Protein::SaveStructure (char structCode, int numAtoms, Point * points, 
			      char * prevLine, char * nextLine ) 
{
  
  Point  vecBase,vecEnd;
  Sse *  retVal;
  int    i;
  float  extraPoint[3];
  //  char record[10];
  static int lastSavedIsLoop = 0;

  retVal = NULL;
  switch(structCode) {
  case 'H': case 'G': case 'I': 
    { // will save a helix structure.
      if (numAtoms < 4) 
	{
	  //cout<< "Not considering the structure as helix since it has " << numAtoms << "amino acids." << endl;
	  //      cout<< "instead saving it as loop.." << endl;
	  retVal = saveAsLoop(&lastSavedIsLoop,points[0], points[numAtoms-1], numAtoms);
      for (i = 0; i< numAtoms; i++) 
	sseTypeSeq.push_back('L');
      if (numAtoms <= 1) {
	cout << "Alert: Saved a loop structure of :" << numAtoms << " atoms instead of a helix. " << endl;
      }
    }
    else if (numAtoms == 4) {
      
      lastSavedIsLoop = 0;
      if (strcmp(prevLine,"")) {
	//	cout << "prevLine = " << prevLine << "length = " << strlen(prevLine) << endl;
	extract(extraPoint,prevLine);
	for (i=0; i<3; i++) {
	  vecBase[i] = (0.74 * extraPoint[i] + points[0][i]  + 
		      points[1][i] + 0.74 * points[2][i]) / 3.48;
	  vecEnd[i]  = (0.74 * points[3][i] + points[2][i]  + 
 			points[1][i] + 0.74 * points[0][i]) / 3.48;
	}
      }
      else {
	if (strcmp(nextLine,"")) {
	  //cout << "Nextline = " << nextLine << endl;
	  extract(extraPoint,nextLine);
	  for (i=0; i<3; i++) {
	    vecBase[i] = (0.74 * points[0][i] + points[1][i]  + 
			  points[2][i] + 0.74 * points[3][i]) / 3.48;
	    vecEnd[i]  = (0.74 * extraPoint[i] + points[3][i]  + 
			  points[2][i] + 0.74 * points[1][i]) / 3.48;
	  }
	}
      }
      retVal          = new Sse(ALPHA, vecBase, vecEnd, numAtoms);
      for (i = 0; i< numAtoms; i++) 
	sseTypeSeq.push_back('H');
    }
    else { // numAtoms > 4
      lastSavedIsLoop = 0;
      //A assert(numAtoms > 4); 
      for (i=0; i<3; i++) {
	vecBase[i] = (0.74 * points[0][i] + points[1][i]  + 
		      points[2][i] + 0.74 * points[3][i]) / 3.48;
	vecEnd[i]  = (0.74 * points[numAtoms - 1][i] + points[numAtoms - 2][i]  + 
		      points[numAtoms - 3][i] + 0.74 * points[numAtoms - 4][i]) / 3.48;
      }
      for (i = 0; i< numAtoms; i++) 
	sseTypeSeq.push_back('H');

      retVal          = new Sse(ALPHA, vecBase, vecEnd, numAtoms);
      //   cout << "Saved an alpha helix of " << numAtoms << " atoms. " << endl;
    }
    break;
  }
  case 'E' : 
    {
    if (numAtoms < 3) {
      //      cout << "Beta sheet has less than 3 atoms."<< endl;//, saving it as loop.." << endl;
      retVal = saveAsLoop(&lastSavedIsLoop,points[0],points[numAtoms - 1], numAtoms);
      for (i = 0; i< numAtoms; i++) 
	sseTypeSeq.push_back('L');
      if (numAtoms <=1) {
	cout << "Alert: Saved a loop of :" << numAtoms << " atoms instead of a helix. " << endl;
      }
    }
    else {
      for (i=0; i<3; i++) {
	vecBase[i] = (points[0][i] + points[1][i]) / 2.0;
	vecEnd [i] = (points[numAtoms - 1][i] + points[numAtoms - 2][i]) / 2.0;
      }
      retVal     = new Sse(BETA,vecBase,vecEnd, numAtoms);
      for (i = 0; i< numAtoms; i++) 
	sseTypeSeq.push_back('E');
      //      cout << "Saved a beta sheet of " << numAtoms << " atoms. " << endl;
      lastSavedIsLoop = 0;
    }
    
    break;
    }
  default : {
    for (i = 0; i< numAtoms; i++) 
      sseTypeSeq.push_back('L');
    retVal = saveAsLoop(&lastSavedIsLoop,points[0], points[numAtoms - 1], numAtoms);
  }
  } //switch
  return retVal;
}
#endif

/* 
   changes ssetypeseq, to make H,G,I H and no sse elements as L.
   also changes the alpha helices of less than 4 elements to loops.
                    beta  strands of less than 3 elements to loops.
*/
void updateSseSeq (vector<char> & sseTypeSeq)
{
  for (unsigned int i = 0; i < sseTypeSeq.size(); i++)
    {
      switch(sseTypeSeq[i])
	{
	case 'H' : case 'G': case 'I':
	  {
	    sseTypeSeq[i] = 'H';
	    break;
	  }
	case 'E' :
	  {
	    // nothing to do
	    break;
	  }
	default :
	  {
	    sseTypeSeq[i] = 'L';
	  }
	}
    }
  /* second pass. to change short sse's to loops. */
  unsigned int i = 0;
  while (i < sseTypeSeq.size())
    {
      switch (sseTypeSeq[i])
	{
	case 'H' : case 'E':
	  {
	    unsigned int j = i + 1;
	    while ((j < sseTypeSeq.size()) && (sseTypeSeq[j] == sseTypeSeq[i]))
	      {
		j++;
	      }

	    // sse   extends from i to j-1.
	    int numAAs = j - i;
	    if (   ((numAAs        < 4)    && (sseTypeSeq[i] == 'H')) 
		   || ((numAAs        < 3)  && (sseTypeSeq[i] == 'E')))
	      {
		cout << "changing sse element extending from ";
		cout << i + 1 << " till " << j << " to L." << endl;
		for (unsigned int k = i; k <= j-1; k++)
		  {
		    sseTypeSeq[k] = 'L';
		  }
	      }
	    i = j - 1; // i will be incremented eventually
	    break;
	  }
	case 'L' :
	  {
	    // nothing to do
	    break;
	  }
	default :
	  {
	    cout << "unrecognized sse element " << sseTypeSeq[i] << ". " << endl;
	    exit(0);
	  }
	}//switch
      i++;
    }
}

/*
  the sse extends from AA#k till AA#l, both included.
  sse type is either H, E or L.
*/
Sse * computeVector (int k, int l, int sseType, 
		     const vector<vector<float> >& cACoords)
{
  Sse * retval;
  int numAAs = l - k + 1;
  assert (numAAs > 0);
  switch (sseType)
    {
    case 'H' :
      {
	if (numAAs < 4) 
	  {
	    retval = loopVector (cACoords[k], cACoords[l], numAAs);
	  }
	else
	  {
	    retval = alphaHelixVector (cACoords, k, l);
	  }
	break;
      }
    case 'E' :
      {
	if (numAAs < 3)
	  {
	    retval = loopVector (cACoords[k], cACoords[l], numAAs);
	  }
	else
	  {
	    retval = betaStrandVector (cACoords, k, l);
	  }
	break;
      }
    case 'L' :
      {
	retval     = loopVector (cACoords[k], cACoords[l], numAAs);
	break;
      }
    default  :
      cout << "sse type " << sseType << " unknown. Exiting..." << endl;
      exit(0);
    }
  return retval;
}


/*
  sse sequence is updated, to contain only H, E or L, for helix, beta strand 
  and loop.
  The vectors of the protein is computed.
  CA coords are in homogeneous coords, first converting them to nonhomogeneous coords.
*/

void Protein::computeSSEVectors(const vector<vector<float> >& cACoords)
{
  assert (cACoords[0].size() == 4);
  vector<vector<float> > newCACoords;

  
  newCACoords.resize(cACoords.size());

  for (unsigned int i = 0; i < cACoords.size(); i++)
    {
      newCACoords[i].resize(3);
      for (unsigned int j = 0; j < 3; j++)
	{
	  newCACoords[i][j] = cACoords[i][j+1];
	}
    }
  
  for (unsigned int i = 0; i < sseTypeSeq.size(); i++)
    {
      cout << sseTypeSeq[i] ;
    }
  cout << endl;
  
  updateSseSeq (sseTypeSeq);
  for (unsigned int i = 0; i < sseTypeSeq.size(); i++)
    {
      cout << sseTypeSeq[i] ;
    }
  cout << endl;

  unsigned int i = 0;
  while (i < sseTypeSeq.size())
    {
      char sseType   = sseTypeSeq[i];
      unsigned int j = i+1;
      while ((j < sseTypeSeq.size()) && (sseTypeSeq[j] == sseType))
	{
	  j++;
	}
      // there is a new sse from sseTypeSeq[i]..sseTypeSeq[j-1].
      Sse * sseVector = computeVector(i, j-1, sseTypeSeq[i], newCACoords);
      //      cout << "computing the vector from AA# " << i+1 << " to " << j ;
      // AA # are printed in 1- base numbers, but the vectors sseTypeSeq,
      // sseVectors are 0-based.
      //      cout << " for a sse of type " << sseTypeSeq[i] << endl;
      sseVectors.push_back (sseVector);
      //      cout << "sse is : ";
      //      cout << (sseVector) << endl;
      i = j;
    }
  computeLoopVectors();
  /*
  cout << "sse vectors are : " << endl;
  for (unsigned int i = 0; i < sseVectors.size(); i++)
    {
      cout << sseVectors[i] << endl;
    }
  */
  computeBondDihedralAngles();
}
/*
  similar to readDSSP, but reads the parsed file, which contains
  .) the number of AAs
  .) aa sequence
  .) sse sequence
 // .) coord info
  
//  cACoords are returned.
*/
void Protein::readParsedDSSPFile (const char * filename)
				  
{
  FILE * infile = fopen(filename,"r");
  
  char ch;
  char tempStr[100];
  int numAAs;

  if (infile == NULL)
    {
      cout << "couldn't open " << filename << endl;
      exit(0);
    }
  fscanf (infile, "%d\n", &numAAs);
  cout << numAAs << " AAs." << endl;
  
  //  fgets (infile, "%s\n", tempStr);
  fgets(tempStr,100,infile);
  //  cout << "sse seq : ";
  for (int i = 0; i < numAAs; i++)
    {
      fscanf (infile, "%c", &ch);
      sseTypeSeq.push_back(ch);
      //cout << sseTypeSeq.back();
    }
  //  cout << endl;
  
  fgetc (infile); // to read the null character
  fgets(tempStr,100,infile);
  //  cout << "AA seq : " ;
  for (int i = 0; i < numAAs; i++)
    {
      fscanf (infile, "%c", &ch);
      sequence.push_back(ch);
      //cout << sequence.back();
    }
  //  cout << endl;
  fgets(tempStr,100,infile);
  
#if 0
  cACoords.resize(numAAs);
  for (int i = 0; i < numAAs; i++)
    {
      cACoords[i].resize(3);
    }
  
  //  fscanf (infile, "%s", tempStr);
  fgets(tempStr,100,infile);
  //cout << "coords : " << endl;
  for (int i = 0; i < numAAs; i++)
    {
      fscanf (infile, "%f %f %f\n", &(cACoords[i][0]), 
	      &(cACoords[i][1]), &(cACoords[i][2]));
    }
#endif
}
#if 0
/* 
   sets:
     AA  sequence
     sse sequence
     the beginning and end positions of the alpha helices, beta strands.
*/

void Protein::readDSSPfile(const char * filename) 
{
  const  int MAXELEMENTS = 100;  /* max. # 2ndary structure elements */
  const  int MAXATOMS    = 1000; /* max. number of c-alpha atoms     */
  FILE * infile;
  char   line[200];
  char   record[10];
  int    i,start, // == 1 when coordinate info starts 
    numAtoms = 0, // number of atoms in the current secondary structure
    structureEnded; // is 1 if last secondary structure ends with a character denoting another 2ndary structure
  
  Sse*   tempElements[MAXELEMENTS];
  char   structureChar; // the character representing 2ndary structure of the current amino acid. 
  Point  points[MAXATOMS];
  int    firstLineRead;
  char   prevLine[200], // holds the line before the structure started.
    tempLine[200]; 
  
  //    ofstream fout ("originalCaLocs.txt");
  if (VERBOSE_MODE) 
    {
      cout << "Make sure the end of the dssp file doesn't contain extra characters." << endl;
      cout << "Also make sure that coordinates are between -999.9 and 9999.9 " << endl;
    }
  
  infile = fopen(filename,"r");
  if (infile == NULL)
    {
      cout << "couldn't open " << filename << endl;
      exit(0);
    }
  start = 0; firstLineRead = 0; strcpy(prevLine, "");

  while ((firstLineRead || 
	  (fgets(line,200,infile))) &&  
	 (numElements <  MAXELEMENTS))
    {
      if (start == 0) 
	{
	  // first finding where the coordinate info starts
	  extStr(record,line,3,12);
	  if (strcmp(record,"#  RESIDUE")  == 0) 
	    {
	      start    = 1;
	      numAtoms = 0;
	    }
	}
      else 
	{
	  firstLineRead = 0; 
	  extStr(record, line, 14, 15);
	  sequence.push_back(record[0]);
	  extStr(record,line,17,18); // not yet in a secondary structure
	  structureChar = record[0];
	  //      cout << record[0];
	  //      saveAPoint(&numAtoms, points, line);
	  extract(points[numAtoms],line);
	  //	    fout << points[numAtoms][0] << " " << points[numAtoms][1] << " " << points[numAtoms][2] << endl;
	  numAtoms++;
	  structureEnded = 0;
	  strcpy(tempLine,line);

	  while ((!structureEnded) && (fgets(line,200,infile))) 
	    {
	      extStr(record,line,17,18);
	      if (record[0] == structureChar) 
		{
		  // structure continues...
		  //	  cout << record[0];
		  extract(points[numAtoms],line);
		  //		    fout << points[numAtoms][0] << " " << points[numAtoms][1] << " " << points[numAtoms][2] << endl;
		  numAtoms++;
		  extStr(record, line, 14, 15);
		  sequence.push_back(record[0]);
		  //	  saveAPoint(&numAtoms, points,line);
		  strcpy(tempLine,line);
		}
	      else 
		{
		  // structure ended.
		  structureEnded = 1;
		}
	    }
	  //      cout << endl;
	    
	  if (!structureEnded) 
	    { // i.e. fgets(line, 200, infile) is false
	      // line is not updated.
	      strcpy(line,"");
	    }
	    
	  
	  tempElements[numElements] = 
	    SaveStructure(structureChar, numAtoms, points,prevLine,line);
	  if (tempElements[numElements] != NULL) 
	    {
	      numElements++;
	    }
	  else 
	    {
	      /* last structure was LOOP, and the previous one too.. Last one is not saved, but the
		 location of the last atom is used to update the loop's end atom coordinate. This is useful if
		 the last vector is a loop, in which case we should keep track of the last atom's coordinates.
		 Also, the number of atoms in the last loop structure is updated. 
	      */
	      assert(numElements > 0);
	      assert(tempElements[numElements - 1]->getType() == LOOP);
	      tempElements[numElements - 1] -> setVector2(tempElements[numElements-1]->getBasePoint(), points[numAtoms - 1]);
	      tempElements[numElements - 1] -> incrementNumAAs(numAtoms);
	    }
	  numAtoms                  = 0;
	    
	  if (structureEnded) {
	    strcpy(prevLine,tempLine);
	    firstLineRead = 1;
	  }
	}
    }
    
  if (numElements >=  MAXELEMENTS) {
    cout << "numAtoms    = " << numAtoms    << " MAXATOMS    = " << MAXATOMS << endl;
    cout << "numElements = " << numElements << " MAXELEMENTS = " << MAXELEMENTS << " . exiting..." << endl;
    exit(1);
  }
  
  //  now save tempElements in the protein data structure, but first check whether an ALPHA-BETA/ BETA-ALPHA structure duo is aligned on a row like this. In that situation, insert a loop in between.
    
  for (i = 0; i< numElements - 1; i++) {
    if ( ((tempElements[i]->getType() == ALPHA) && (tempElements[i+1]->getType() == BETA)) || 
	 ((tempElements[i]->getType() == BETA)  && (tempElements[i+1]->getType() == ALPHA))) {
      numElements++;
      cout << "Inserted a loop as the " << i+1 << " th vector. " << endl;
      for (int j = numElements; j >= (i+2); j--) {
	tempElements[j] = tempElements[j-1];
      }
      tempElements[i+1] = new Sse(LOOP, points[0], points[0],0); // just giving it n'importe quel point. 
      //Also note that the number of atoms is 0.
    }
  }
  elements = new pSse[numElements];
    
  for (i = 0; i< numElements; i++) {
    elements [i] = tempElements[i];
  }
    
  if (VERBOSE_MODE) 
    {
      cout << "Here is the amino acid sequence. " ;
      int sequenceSize = sequence.size();
      for (i = 0; i< sequenceSize; i++) 
	{
	  cout << sequence[i] ;
	}
      cout << endl << "total " << sequenceSize << " amino acids. " << endl;
      cout << "SSE sequence : " ;
      int sseTypeSeqSize = sseTypeSeq.size();
      for (i = 0; i< sseTypeSeqSize; i++) 
	{
	  cout << sseTypeSeq[i] ;
	}
    }
  //    cout << endl << "total " << sseTypeSeq.size() << " sses. " << endl;
    
  computeLoopVectors();
  computeBondDihedralAngles();
  //  findParams(); parameters will be found after the standardization, in 
  // initnative structure
}
#endif

/* different from readrel.centroidpositions in that even for loops the exact value is read. */
void Protein::readRelativecAlphaPositions(const char filename[]) 

{
  ifstream fin(filename);
  
  if (!fin) {
    cout << filename << " was not opened. " << endl;
    exit(0);
  }
  else
    {
      cout << "reading " << filename << endl;
    }
  
  int i = 0;
  int sequenceSize = sequence.size();
  while ((!fin.eof()) && (i<sequenceSize)) {
    fin >> relcAlphaCoords[i][0] >> relcAlphaCoords[i][1] >> relcAlphaCoords[i][2];
    i++;
    if (fin.eof())
      break;
  }
  assert(i == sequenceSize );
}

/* charged and charge arrays filled in here. */
void Protein::setCharges() {

  
  unsigned int i, sequenceSize;
  

  sequenceSize = sequence.size();
#ifdef DEBUG
  assert(charged.size() == 0);
  assert(charge.size()  == 0);
#endif
  for (i = 0;  i < sequenceSize; i++) {
    sequence[i] = toupper(sequence[i]);
    switch(sequence[i]) {
    case 'D': case 'E' : {
      charged.push_back(true);
      charge.push_back(-1.0);
      break;
    }
    case 'K': case 'R' : {
      charged.push_back(true);
      charge.push_back(+1.0);
      break;
    }
    default : {
      charged.push_back(false);
      charge.push_back(0.0) ;
    }}
  }

  assert(charged.size() == charge.size());
  assert(charged.size() == sequenceSize);


}
/* 
   allocates, and reads c-a coords, centroid coords, C', N and O 
   coordinates.
   set the charge of some amino acids.
   sets loop array, which contain true for every amino acid in a loop.
   sets hydrophobicity array, which contains true for every
        hydrophobic amino acid.
	
   the last character of path2DataFiles is "/".
*/

void Protein::doSetup(const char * path2DataFiles)
{
  int  i, sequenceSize;
  char tempStr[100];//, header[100];
  char tempStr2[100], tempStr3[100];

  sequenceSize = sequence.size();
  relcAlphaCoords  = new pFloat[sequenceSize];
  for (i = 0; i<sequenceSize ; i++) 
    {
      relcAlphaCoords[i] = new float[3];
    }
  
  relCentroidCoords = new pFloat[sequenceSize];

  for (i = 0; i<sequenceSize ; i++) 
    {
      relCentroidCoords[i] = new float[3];
    }

  NUMDOF = (numElements - 1) * 3;

  cout << "NUM DOF = " << NUMDOF << endl;
  //  setHPMapping();
  
  assert (path2DataFiles[strlen(path2DataFiles) - 1] == '/');

  strcpy(tempStr,  path2DataFiles);
  strcat(tempStr, "cAlphaRelDict.txt");
  readRelativecAlphaPositions    (tempStr);

  strcpy(tempStr,  path2DataFiles);
  strcat(tempStr, "centroidRelDict.txt");
  readRelativeCentroidPositions  (tempStr);
  
  strcpy(tempStr,  path2DataFiles);
  strcat(tempStr, "CprimeRelDict.txt");

  strcpy(tempStr2, path2DataFiles);
  strcat(tempStr2, "NRelDict.txt");

  strcpy(tempStr3, path2DataFiles);
  strcat(tempStr3, "ORelDict.txt");
 

  readRelativeCprime_N_OPositions(tempStr, tempStr2, tempStr3);
  setCharges();
  setHydrophobicity();
  setLoops();

  for (i = 0; i< sequenceSize; i++) 
    {
      sequence[i] = toupper(sequence[i]);
    }

  int sseTypeSeqSize = sseTypeSeq.size();
  for (i = 0; i< sseTypeSeqSize; i++) 
    {
      sseTypeSeq[i] = toupper(sseTypeSeq[i]);
    }
}




/* 
   sets the loop vector. 
   loop[i] == true if i'th amino acid's
   sseType is 'L'.
   not discarding any amino acid, i.e. first and last vectors are also counted.
*/
void Protein::setLoops() 
{
  for (unsigned int AAindex = 0; AAindex < sseTypeSeq.size(); AAindex ++) 
    {
      if (sseTypeSeq[AAindex] == 'L') 
	{
	  loop[AAindex] = true;
	}
      else
	{
	  loop[AAindex] = false;
	  nonLoopAAIndices.push_back(AAindex);
	}
    }
}

#if 0
/*
  does the shrinking of the vectors, in case first and/or 
  last vectors are LOOP. 
  sets loop vectors between prev. vector end point and next vector 
  start point.
*/

void Protein::computeLoopVectors() {

  int i;

  lastAAIndexOfLastNonLoopVector_plusOne = getNumAAs();

  if (elements[numElements - 1] ->getType() == LOOP) 
    {
      lastAAIndexOfLastNonLoopVector_plusOne -= elements[numElements - 1]->getNumAAs();
      delete elements[numElements - 1];
      elements[numElements - 1] = NULL;
      numElements --;
    }
  
  if (elements[0] ->getType() == LOOP) 
    {
      firstAAIndexOfFirstNonLoopVector       += elements[0]->getNumAAs();
      delete elements[0];
      elements[0] = NULL;
      numElements--;
    }
  
  cout << " firstAAIndex  = " <<  firstAAIndexOfFirstNonLoopVector  << " lastAAIndex = " << lastAAIndexOfLastNonLoopVector_plusOne << endl;
  
  if (!elements[0]) 
    {
      for (i = 0; i < numElements; i++) 
	{
	  elements[i] = elements[i+1];
	}
      elements[numElements] = NULL;  // so that in the destructor, elements[numElements - 1] will be deleted only once
    }
  
  for (i=0; i<Protein::numElements; i++)  
    if (elements[i]->type == LOOP) 
      {
	assert ((i >0) && ( i < numElements -1));
	elements[i]->setVector2(elements[i-1]->vec.getEnd(), 
				elements[i+1]->getBasePoint());
	
      }

}
#endif

/*
  does the shrinking of the vectors, in case first and/or 
  last vectors are LOOP. 
  sets loop vectors between prev. vector end point and next vector 
  start point.
  sets lastAAIndexOfLastNonLoopVector_plusOne and 
       firstAAIndexOfFirstNonLoopVector.
*/

void Protein::computeLoopVectors() 
{
  lastAAIndexOfLastNonLoopVector_plusOne = getNumAAs();
  
  Sse * lastSSE = sseVectors.back();
  assert (lastSSE != NULL);
  
  vector<Sse *> debugSSEPtrVector;

  // this vector is to check for whether the sse elements, when 
  // the vector sseVector is resized properly, still keep the
  // right elements.

  for (unsigned int i = 0; i < sseVectors.size(); i++)
    {
      debugSSEPtrVector.push_back (sseVectors[i]);
    }
  
  if (lastSSE->getType() == LOOP)
    {
      lastAAIndexOfLastNonLoopVector_plusOne -= lastSSE->getNumAAs();
      delete lastSSE;
      sseVectors.resize(sseVectors.size() - 1);
      for (unsigned int i = 0; i < sseVectors.size(); i++)
	{
	  assert (sseVectors[i] == debugSSEPtrVector[i]);
	}
  //  cout << "checking whether resize keeps the prev. elements when a resize";
  //  cout << " is done." << endl;
    }
  
  Sse * firstSse = sseVectors.front();

  if (firstSse ->getType() == LOOP) 
    {
      firstAAIndexOfFirstNonLoopVector  += firstSse->getNumAAs();
      delete firstSse;
      sseVectors.front() = NULL;
      
      for (unsigned int i = 0; i < sseVectors.size() - 1; i++) 
	{
	  sseVectors[i] = sseVectors[i+1];
	}
      sseVectors.resize(sseVectors.size() - 1);
       for (unsigned int i = 0; i < sseVectors.size(); i++)
	{
	  assert (sseVectors[i] == debugSSEPtrVector[i+1]);
	}
       //       cout << "assert of sses being resized correctly passed." << endl;
    }
  
  //  cout << " firstAAIndex  = " <<  firstAAIndexOfFirstNonLoopVector  << " lastAAIndex = " << lastAAIndexOfLastNonLoopVector_plusOne << endl;
  
  for (unsigned int i = 0; i < sseVectors.size(); i++)
    {
      Sse * ptr2Sse = sseVectors[i];
      if (ptr2Sse->type == LOOP) 
	{
	  assert ((i > 0) && (i <  sseVectors.size()-1));

	  Sse * prevSse = sseVectors[i-1];
	  Sse * nextSse = sseVectors[i+1];

	  ptr2Sse->setVector2(prevSse->vec.getEnd(),
			      nextSse->getBasePoint());
	}
    }
  Protein::numElements = sseVectors.size();

  cout << "# AAs discarded at the end of the protein       : " ;
  cout << (getNumAAs() - lastAAIndexOfLastNonLoopVector_plusOne) << endl;
  cout << "# AAs discarded at the beginning of the protein : " ;
  cout << firstAAIndexOfFirstNonLoopVector << endl;
}

ostream& operator<< (ostream & output, const Protein& p) {

  int i;

  output << "Printing protein..." << endl;
  output << "num sse elements = " << p.sseVectors.size() << endl;

  for (i = 0; i < p.numElements; i++) 
    {
      output << p.sseVectors[i];
    }
  output << "energy = " << p.getEnergy() << endl;
  //  output << "rmsd from native : " << p.getRMSDFromNative() << endl;
   
  return output;
}

/* 
   copies : 
      .) the sse info (type, base and end point, # AAs, vector angle, 
      dihedral angle, and twist angle.).
      .) centroid coords, and cAlpha Coords.
      .) parameters of the protein.
*/

void Protein::copyConfig(const Protein * original) 
{
  int i,
    arraySize; // the size of the centroidCoords and cAlphaCoords Arrays.
  
  arraySize       = lastAAIndexOfLastNonLoopVector_plusOne -  
    firstAAIndexOfFirstNonLoopVector;
  
  for (i = 0; i< numElements; i++) 
    {
      Sse * ptr2Sse = new Sse (original -> sseVectors[i] -> getType(), 
			       original -> sseVectors[i] -> getBasePoint(), 
			       original -> sseVectors[i] -> getEndPoint(), 
			       original -> sseVectors[i] -> getNumAAs());
      
      sseVectors.push_back (ptr2Sse);
      
      sseVectors[i]->setVectorAngle   
	(original->sseVectors[i]->getVectorAngle());
      sseVectors[i]->setDihedralAngle 
	(original->sseVectors[i]->getDihedralAngle());
      sseVectors[i]->setTwistAngle    
	(original->sseVectors[i]->getTwistAngle());
    }
  
  if (original->centroidCoords == NULL)
    centroidCoords = NULL;
  else 
    {
      centroidCoords = new pFloat [arraySize];
      for (i = 0; i< arraySize; i++) 
	{
	  centroidCoords[i] = new float[3];
	}
      for (i = 0; i< arraySize; i++) 
	{
	  VecCopy(centroidCoords[i], original      ->centroidCoords[i]);
	}
    }
    
  if (original->cAlphaCoords == NULL)
    cAlphaCoords = NULL;
  else 
    {
      cAlphaCoords = new pFloat [arraySize];
      
      for (i = 0; i< arraySize; i++) 
	{
	  cAlphaCoords[i] = new float[3];
	}

      for (i = 0; i< arraySize; i++) 
	{
	  VecCopy(cAlphaCoords[i], original->cAlphaCoords[i]);
	}
    }

  energy   = original->energy;
  myNode   = NULL; myTree = NULL;
  numLinks = 0;
  for (unsigned int k = 0; k < original->params.size(); k++)
    {
      params.push_back(original->params[k]);
    }
}

/* 
   finds a random theta and phi for alpha helices and beta sheets, and keeps their length same. 
   finds the base and the tip of the vectors using the previous vectors, if any, and the angles. Set vector length accordingly.
   finds a random length around the original length (between 20% and 180% of the original length) and keeps angles same for loops. 
   finds the base and the tip of the vector from the previous vector, the new length of the vector and the angles theta and phi. 
  assumes the elements data structure is NULL.   
  update : also finds twistangle for alpha helices and beta sheets, between 0 and 2 * pi.
 
*/

void Protein::createRandomConfig(const Protein * nativeP) 
{
  int i;
  float vectorAngle, dihedralAngle, twistAngle;
  const Point originPoint = {0.0, 0.0, 0.0};

  //A  assert(elements == NULL);
  
  for (i = 0; i< numElements; i++) 
    {
      Sse * ptr2Sse = new Sse (nativeP->sseVectors[i]->getType(), 
			       originPoint, 
			       originPoint, 
			       nativeP->sseVectors[i]->getNumAAs());
      
      sseVectors.push_back (ptr2Sse);
      assert(sseVectors[i]  != NULL);
      
      switch (sseVectors[i]->getType())
	{
	case ALPHA :
	case BETA : 
	  {
	    
	    vectorAngle       = (((float)(random()))/((float)(MY_RAND_MAX))) *
	      M_PI;
	    dihedralAngle     = (2 * (((float)(random()))/((float)(MY_RAND_MAX)))- 1.0) * M_PI;
	    twistAngle        = (2 * (((float)(random()))/((float)(MY_RAND_MAX))) - 1.0) * M_PI;
	    //      cout << "generated twist Angle in create random config = " << twistAngle << endl;
	    switch(i) 
	      {
	      case 0 : 
		{
		  sseVectors[0]->computeVector 
		    (originPoint, nativeP->sseVectors[0]->getLength(), 
		     vectorAngle, dihedralAngle, NULL, NULL);
		  // base at the origin 
		  break;
		}
	      case 1 : 
		{
		  sseVectors[1]->computeVector (sseVectors[0] -> getEndPoint(), nativeP->sseVectors[1]->getLength(), vectorAngle, dihedralAngle, sseVectors[0], NULL);
		  break;
		}
	      default: 
		{
		  sseVectors[i]->computeVector(sseVectors[i-1]-> getEndPoint(),  nativeP->sseVectors[i]->getLength(), vectorAngle, dihedralAngle, sseVectors[i-1], sseVectors[i-2]);
		}
	      } // switch
	    sseVectors[i]->setTwistAngle(twistAngle);
	    break;
	  }
	case LOOP : 
	  {
	    /* 3 dof for loops with AAs, 2 DOF for loops w/o AAs 
	       (those loops that are inserted between 2 consecutive alpha/beta - alpha/beta)
	    */
	    float length, randomNumber;
	    randomNumber = (((float)(random()))/((float)(MY_RAND_MAX)));
	    int numAAs   = nativeP->sseVectors[i]->getNumAAs();
	    if (numAAs > 0) {
	      length   = numAAs * (Protein::minLengthPerAA + randomNumber * (Protein::maxLengthPerAA - Protein::minLengthPerAA));
	    }
	    else {
	      length   = nativeP->sseVectors[i] ->getLength(); /*but the length of the vector if loop has no AA
							       is constant! Hence 2 DOF!! */
	    }
	    
	    vectorAngle      =  (((float)(random()))     /((float)(MY_RAND_MAX)))        * M_PI;
	    dihedralAngle  =  (2 * (((float)(random()))/((float)(MY_RAND_MAX))) - 1.0) * M_PI;
	    twistAngle     =  0;
	    
	    sseVectors[i]->setTwistAngle(twistAngle);
	    
	    switch(i) 
	      {
	      case 0 : {
	      sseVectors[0]->computeVector(originPoint,                length, vectorAngle, dihedralAngle, NULL, NULL);
	      break;
	    }
	    case 1 : {
	      sseVectors[1]->computeVector(sseVectors[0]->getEndPoint(), length, vectorAngle, dihedralAngle, sseVectors[0], NULL);
	      break;
	    }
	    default : {
	      sseVectors[i]->computeVector(sseVectors[i-1]->getEndPoint(), length, vectorAngle, dihedralAngle, sseVectors[i-1], sseVectors[i-2]);
	    }
	    }
	    break;
	  }
	default: {
	    cout << "No such secondary structure element!" << endl;
	    exit(1);
	  }
	}// switch
    }

  computeEnergy();
  findParams();
  //  setRMSDField(nativeP);
}

/* same as protein::protein, with the same input */

void    Protein::createGivenConfig(const vector<float> &dih, 
				   const vector<float> &vec, 
				   const vector<float> &twi, 
				   const vector<float> &len,
				   const Protein * nativeP)
{
  
  const Point originPoint = {0.0, 0.0, 0.0};
  
  assert(numElements == 7); // for 256B only for now.
  
  for (int i = 0; i< numElements; i++) 
    {

      Sse * ptr2Sse = new Sse (nativeP->sseVectors[i]->getType(), originPoint, 
			       originPoint, nativeP->sseVectors[i]->getNumAAs());
      sseVectors.push_back (ptr2Sse);
      sseVectors[i]->setTwistAngle(twi[i]);
      switch(i) {
      case 0 : {
	sseVectors[0]->computeVector(originPoint,                len[i], vec[i], dih[i], NULL, NULL);
	break;
      }
      case 1 : {
	sseVectors[1]->computeVector(sseVectors[0]->getEndPoint(), len[i], vec[i], dih[i], sseVectors[0], NULL);
	break;
      }
      default: {
	sseVectors[i]->computeVector(sseVectors[i-1]->getEndPoint(), len[i], vec[i], dih[i], sseVectors[i-1], sseVectors[i-2]);
	
      }
      } // switch
      
    }
  computeEnergy();
  findParams();
  //  setRMSDField(nativeP);

}

// computes the energy.

float Protein::computeEnergy() 
{
  //  generateCAlphaCoords();
  generateCentroidCoords();
#if INCLUDE_HYDROGEN_BONDS
  generateCprime_N_O_H_Coords_InvolvedInHBonds ();
#endif
  _computeEnergy(); // energy field should be set here.
  
  return energy;
}


/* note: normally resize should also initialize, but I couldn't get it working, maybe since because this is a multidimensional vector */

#if 0



void Protein::clearGrid() {
  
  int i,j,k;
  int maxDisplacement = rint(AApairDistance / gridResolution); // max displacement in the grid in one way
  int _2MD_1 = 2 * maxDisplacement + 1;
  int sizeX, sizeY, sizeZ;
  
  sizeX = ((max[0]-min[0]) / (gridResolution)) +  _2MD_1;
  sizeY = ((max[1]-min[1]) / (gridResolution)) +  _2MD_1;
  sizeZ = ((max[2]-min[2]) / (gridResolution)) +  _2MD_1;
  
  grid.resize(sizeX);
  for (i = 0; i < grid.size(); i++) {
    grid[i].resize(sizeY);
    for ( j = 0; j < grid[i].size(); j++) {
      grid[i][j].resize(sizeZ);
    }
  }

  cout << "Grid. cleared and resized. Now, sizes are : " << sizeX << " " << sizeY << " " << sizeZ << " " << endl;
  assert(grid.size() == sizeX);
  for (i = 0; i< grid.size(); i++)
    for (j = 0; j< grid[i].size(); j++) 
      for (k = 0; k< grid[i][j].size(); k++) {
	grid[i][j][k] = false;
      }
}
#endif

#if 0
float Protein::searchAround( int secondaryStructureIndex, int aaIndex, float * minDistance) {
  int xO, yO, zO;
  int x,y,z,i;
  float score = 0.0;
  
  int maxDisplacement = rint(AApairDistance / gridResolution); // max displacement in the grid in one way
  xO = gridCoordinates[secondaryStructureIndex][aaIndex][0];
  yO = gridCoordinates[secondaryStructureIndex][aaIndex][1];
  zO = gridCoordinates[secondaryStructureIndex][aaIndex][2];
  
  if (grid[xO][yO][zO] == false) {
    cout << "xO = " << xO << "yO = " << yO << " zO = " << zO << endl;
    cout << "secondarStructureIndex = " << secondaryStructureIndex << " aaIndex = " << aaIndex << endl;
  }
  assert(grid[xO][yO][zO] == true);
  
  int numGridPointsVisited = 0;
  for (x = xO - maxDisplacement; x <= xO + maxDisplacement; x++) 
    for (y = yO - maxDisplacement + (int) fabs(x - xO); y <=  yO + maxDisplacement - (int) fabs(x-xO); y++) 
      for (z = zO - maxDisplacement + (int) fabs (x - xO) + (int) fabs(y-yO); z <= zO + maxDisplacement - (int) fabs(x - xO) - (int) fabs( y - yO); z++) {
	numGridPointsVisited++;
	if (grid[x][y][z]) {
	  score += 1.0;
	  float tempDistance = (x - xO)*(x - xO) + (y - yO) * (y - yO) + (z - zO) * (z - zO);
	  if ((tempDistance < *minDistance) && (x != xO) && (y != yO) && (z != zO)) {
	    *minDistance = tempDistance;
	  }
	}
      }
  //  cout << numGridPointsVisited << " grid points are visited for each hydrophilic amino acid. " << endl;
  return score;
}
#endif
float Protein::getEnergy() const {

  return energy;
}

float   Protein::getExclusionEnergy()   const {
  return exclusionEnergy;
}

int Protein::getNumLinks() const {

  //  return neighbours.size();
  return numLinks;
}
/*
const  neighbourType& Protein::getNeighbour(int i) const {
  return neighbours[i];
}
*/

// returns # sse elements
int Protein::getNumSseElements()  {
  return numElements;
}


Sse *  Protein::getSseElement(int i) const 
{        // access i'th secondary structure element.
  return sseVectors[i];
}

int Protein::getNumAAs() 
{
  return sequence.size();
}

void Protein::readRelativeCentroidPositions(const char filename[]) 
{
  /* same as readRelativecAlphaPositions, except that
     there are a fixed # of entries (20)
  */
  ifstream fin(filename);
  
  if (!fin) 
    {
      cout << filename << " was not opened. " << endl;
      exit(0);
    }
  else
    {
      cout << "reading " << filename << "..." << endl;
    }

  int i = 0;
  int sequenceSize = sequence.size();
  while ((!fin.eof()) && (i<sequenceSize)) {
    fin >> relCentroidCoords[i][0] >> relCentroidCoords[i][1] >> relCentroidCoords[i][2];
    if (loop[i]) {
      relCentroidCoords[i][0] = relCentroidCoords[i][1] = relCentroidCoords[i][2] = -1.0; 
    }
    i++;
    if (fin.eof())
      break;
  }
  assert(i == sequenceSize);
}

/* 
   i (and j) is the i'th centroid, starting to count from the first centroid of the first non-loop vector. 
*/
float Protein::distanceSquaredBetweenCentroids(int i, int j) 
{
  
  float vect[3], retval;

  //A  assert(centroidCoords != NULL);
  VecSub(vect, centroidCoords[i], centroidCoords[j]);
  retval = vect[0]  * vect[0] + vect[1]  * vect[1] + vect[2]  * vect[2] ;
  return retval;
}

/* i and j are defined as above. */
float Protein::distanceSquaredBetweenCAlphas  (int i,int j) {

  float vect[3], retval;
  assert(cAlphaCoords != NULL);
  
  
  VecSub(vect, cAlphaCoords[i], cAlphaCoords[j]);
  retval = vect[0]  * vect[0] + vect[1]  * vect[1] + vect[2]  * vect[2] ;
  return retval;
}


// returns the alphabetic rank of the aa. 
int findIndex(const char AAchar) {
  // very similar to findAAnumber in genSCData.C
  static char shortAAnames[20];
  static int setupDone = 0;
  int i, index;
  
  if (!setupDone) {

    shortAAnames[0]  = 'A';
    shortAAnames[1]  = 'C';
    shortAAnames[2]  = 'D';
    shortAAnames[3]  = 'E';
    shortAAnames[4]  = 'F';
    shortAAnames[5]  = 'G';
    shortAAnames[6]  = 'H';
    shortAAnames[7]  = 'I';
    shortAAnames[8]  = 'K';
    shortAAnames[9]  = 'L';
    shortAAnames[10] = 'M';
    shortAAnames[11] = 'N';
    shortAAnames[12] = 'P';
    shortAAnames[13] = 'Q';
    shortAAnames[14] = 'R';
    shortAAnames[15] = 'S';
    shortAAnames[16] = 'T';
    shortAAnames[17] = 'V';
    shortAAnames[18] = 'Y';
    shortAAnames[19] = 'W';
    setupDone  =  1;
  }
  
  index = -1;
  for (i=  0; i< 20; i++) {
    if (shortAAnames[i] ==  AAchar) {
      index=  i;
      break;
    }
  }

  if (index  == -1) {
    cout << "Couldn't find AA code. AA code was :" << AAchar << endl;
  }
  
  return index;
}
/*
void Protein::addNeighbour(neighbourType neigh) {
  neighbours.push_back(neigh);
}
*/

void Protein::setMyNode (nodeClass * node) {     // sets the myNode variable to the input parameter
  myNode = node;

}

// returns true if ith amino acid is in loop structure. i = 0.. numAAs-1
#if 0
bool   Protein::isLoop(int AAindex) {

  assert((AAindex >=0) && (AAindex < sseTypeSeq.size()));

  if (sseTypeSeq[AAindex]=='L') {
    return true;
  }
  else  {
    return false;
  }
}
#endif

void Protein::setEnergy(float iEnergy) {
    energy = iEnergy;

}

// returns true if configuration does not collide with itself, and is low than a probabilistically generated energy threshold.
// maxExEnergy is used as the max. energy a node can have. maxExEnergy should
// be > minEnergy.
bool Protein::checkCollision() {

  if (exclusionEnergy > Protein::maxExEnergy) {
    return false;
  }
  else {


    float randomNumber        = (((float)(random()))/((float)(MY_RAND_MAX)));
    float acceptanceThreshold = (energy - Protein::minEnergy) / 
      (Protein::maxExEnergy - Protein::minEnergy);
    if (acceptanceThreshold < randomNumber) {
      return true;
    }
    else { 
      return false;
    }

  }
}

void generateValueAroundOriginal(float * retVal, float original, float increment) {
  float randomNumber;

  //  cout << "value passed = " << original << endl;
  randomNumber = (((float)(random()))/((float)(MY_RAND_MAX)));
  (*retVal)    = original + randomNumber *(2.0 * increment) - increment;
}
#if 0
float Protein::getRMSDFromNative() const {
  testInvariant();
  return rmsdFromNative;
}
#endif

// creates a config. around the original.
void Protein::createNonRandomConfig(const Protein * original, 
				    Protein ** allNodes) 
{
    
  int   i,numelements, numAAs;
  float length, twistAngle, vectorAngle, dihedralAngle,minLength,maxLength;
  const Point originPoint = {0.0, 0.0, 0.0};

 
  //  assert(elements == NULL);
  
  numelements     = original->numElements;
  //  elements        = new pSse[numElements];
  
  for (i = 0; i< numElements; i++) 
    {

      numAAs      =  original->sseVectors[i]->getNumAAs();
      Sse * ptr2Sse = new 
	Sse (original->sseVectors[i]->getType(), originPoint, 
	     originPoint,numAAs);

      sseVectors.push_back (ptr2Sse);
      
      //      cout << "printing " << i << " th sse." << endl;
      //      cout << elements[i] << endl;

      /* 
	 cout << "i = " << i << " original numAAs = " << 
	 original->sseVectors[i]->getNumAAs() << " this one has : " ;
	 cout << sseVectors[i]->getNumAAs() << endl;
      */
      switch(sseVectors[i]->getType()) 
	{
	case ALPHA : 
	case BETA : 
	  {
	    //	    cout << "angle passed = " << original->sseVectors[i]->getVectorAngle() << endl;
	    generateValueAroundOriginal
	      (&vectorAngle,     original->sseVectors[i]->getVectorAngle(), 
	       Protein::angle_epsilon);
	    generateValueAroundOriginal
	      (&dihedralAngle, original->sseVectors[i]->getDihedralAngle(), 
	       Protein::angle_epsilon);
	    generateValueAroundOriginal
	      (&twistAngle,    original->sseVectors[i]->getTwistAngle(), 
	       Protein::angle_epsilon);

	    if (vectorAngle > M_PI) 
	      vectorAngle = 2 * M_PI - vectorAngle;
	    else if (vectorAngle < 0) 
	      vectorAngle = - vectorAngle;
	    assert((vectorAngle >= 0) && (vectorAngle < M_PI));
	    
	    if (dihedralAngle > M_PI)
	      dihedralAngle   = - 2 * M_PI + dihedralAngle;
	    else if (dihedralAngle <= -M_PI)
	      dihedralAngle   =   2 * M_PI + dihedralAngle;
	    assert((dihedralAngle > -M_PI) && ( dihedralAngle <= M_PI));

	    if (twistAngle > M_PI)
	      twistAngle   = - 2 * M_PI + twistAngle;
	    else if (twistAngle <= -M_PI)
	      twistAngle   =   2 * M_PI + twistAngle;
	    assert((twistAngle > -M_PI) && ( twistAngle <= M_PI));

	    sseVectors[i]->setTwistAngle(twistAngle);

	    switch(i) 
	      {
	      case 0 : 
		{
		  sseVectors[i]->computeVector
		    (originPoint, original->sseVectors[0]->getLength(), vectorAngle, dihedralAngle, NULL, NULL);
		  break;
		}
	      case 1 : 
		{
		  sseVectors[i]->computeVector
		    (sseVectors[0]-> getEndPoint(),  original->sseVectors[1]->getLength(), vectorAngle, dihedralAngle, 
		     sseVectors[0], NULL);
		  break;
		}
	      default : 
		{
		  sseVectors[i]->computeVector
		    (sseVectors[i-1]-> getEndPoint(),  original->sseVectors[i]->getLength(), vectorAngle, dihedralAngle, 
		     sseVectors[i-1], sseVectors[i-2]);
		}
	      }//switch
	    break;
	  }
	case LOOP : 
	  {
	    // independent of the # AAs, I perturb the length in the same amount.
	    generateValueAroundOriginal(&length ,   
					original->sseVectors[i]->getLength(),  
					Protein::length_epsilon);
	    generateValueAroundOriginal(&vectorAngle, 
					original->sseVectors[i]->getVectorAngle()
					, Protein::angle_epsilon);
	    generateValueAroundOriginal(&dihedralAngle, 
					original->sseVectors[i]->getDihedralAngle(), 
					Protein::angle_epsilon);

	    minLength = Protein::minLengthPerAA * 
	      original->sseVectors[i]->getNumAAs();
      
	    if (length < minLength) 
	      {
		length = minLength;
	      }

	    else if (length > 
		     (maxLength = 
		      Protein::maxLengthPerAA *original->sseVectors[i]->getNumAAs())) {
	      length = maxLength;
	    }
      
      

	    if (vectorAngle > M_PI)
	      vectorAngle = 2 * M_PI - vectorAngle;
	    else if (vectorAngle < 0)
	      vectorAngle = - vectorAngle;
	    assert((vectorAngle >= 0) && (vectorAngle < M_PI));

	    if (dihedralAngle > M_PI)
	      dihedralAngle   = - 2 * M_PI + dihedralAngle;
	    else if (dihedralAngle <= -M_PI)
	      dihedralAngle   =   2 * M_PI + dihedralAngle;
	    assert((dihedralAngle > -M_PI) && ( dihedralAngle < M_PI));

	    switch(i) 
	      {
	      case 0 : 
		{
		  sseVectors[i]->computeVector(originPoint, length, 
					     vectorAngle, dihedralAngle, 
					     NULL, NULL);
		  break;
		}
	      case 1 : 
		{
		  sseVectors[i]->computeVector(sseVectors[0]->getEndPoint(), 
					     length, vectorAngle, 
					     dihedralAngle, sseVectors[0], NULL);
		  break;
		}
	      default: 
		{
		  sseVectors[i]->computeVector(sseVectors[i-1]->getEndPoint(), 
					     length, vectorAngle, 
					     dihedralAngle, sseVectors[i-1], 
					     sseVectors[i-2]);
		}
	      }//switch

	    twistAngle = 0;
	    sseVectors[i]->setTwistAngle(twistAngle);
	    break;
	  }
	default: 
	  {
	    cout << "No such secondary structure element!" << endl;
	    exit(1);
	  }
	}// switch
    }
  computeEnergy();
  //  setRMSDField(allNodes[0]);
  findParams();
    
}

// currentSseIndex == 0 .. numSseElements -1.
void findBaseAndUnitVecsForTheWholeSse (float base[], float unitX[], 
					float unitY[], float unitZ[], 
					const Protein *myProtein, 
					int currentSseIndex) 
{
  const float epsilon = 1E-6, biggerEpsilon = 1E-4; 
  float vecLen, vecLenSqr;

  assert(myProtein != NULL);
  assert((currentSseIndex >=0) && 
	 (currentSseIndex < myProtein->getNumSseElements()));
  
  Sse * currentSse = myProtein->getSseElement(currentSseIndex);

  VecCopy(base, currentSse->getBasePoint());
  VecSub (unitZ, currentSse->getEndPoint(), currentSse->getBasePoint());
  float vecLength =  currentSse->getLength();
  assert(vecLength != 0.0);
  unitZ[0] /= vecLength; unitZ[1] /= vecLength; unitZ[2] /= vecLength;
  assert(fabs(VecLength(unitZ) - 1) <= biggerEpsilon);

#if 0
  /* this section is for debugging */
  {
    if (
	cout << (*myProtein);
	cout << "debug: currentSseIndex = " << currentSseIndex << " base : " << base[0] << " " <<  base[1] << " " << base[2] << endl;
	cout << "veclength = " << vecLength << " endPoint : " << currentSse->getEndPoint()[0] << " " << currentSse->getEndPoint()[1] << " " << currentSse->getEndPoint()[2] << endl;
	}
    /* end debugging section */
#endif

    vecLenSqr = unitZ[0] * unitZ[0] + unitZ[1] * unitZ[1];
    if (vecLenSqr < epsilon) {
      unitY[1] = 1.0; unitY[0] = unitY[2] = 0.0;
    }
    else {
      unitY[0] = -unitZ[1]; unitY[1] = unitZ[0]; unitY[2] = 0;

      vecLen = sqrt(vecLenSqr);
      for (int i = 0; i < 3; i++) 
	unitY[i] /= vecLen;
    }
   
    VecCross(unitX, unitY, unitZ);
    if  (!(fabs(VecLength(unitX) - 1.0) < biggerEpsilon)) {
      cout << (*myProtein);
      cout << "debug: currentSseIndex = " << currentSseIndex << " base : " << base[0] << " " <<  base[1] << " " << base[2] << endl;
      cout << "veclength = " << vecLength << " endPoint : " << currentSse->getEndPoint()[0] << " " << currentSse->getEndPoint()[1] << " " << currentSse->getEndPoint()[2] << endl;
      cout << "unitX : " << unitX[0] << " " << unitX[1] << " " << unitX[2] << endl;
      cout << "unitY : " << unitY[0] << " " << unitY[1] << " " << unitY[2] << endl;
      cout << "unitZ : " << unitZ[0] << " " << unitZ[1] << " " << unitZ[2] << endl;
    }
    assert (fabs(VecLength(unitX) - 1.0) < biggerEpsilon);  
  }



/* 
  all cAlpha locs (except the ones on the first and/or last vectors, 
  if they are loops)
  are computed. (i.e. no -1 for AAs in loops). 
  The same twist angle of the SSE vector used in centroid computation also 
  used here to find phi of the cAlpha.
  Also, if the C-alphas are already computed, immediately returns, 
  and does not complain.
*/

void Protein::generateCAlphaCoords() 
{

    int i, j, numCAlphas, 
      currentSseIndex,        // 0 .. # sseElements - 1
      relAAIndex ;            // 0.. numAAs in currentSseElement - 1
    float base[3],            // base of the vector corresponding to AAindex
      unitX[3],               /* unit vec. corresponding to
				 each amino acid in the current sse.
			      */
      unitY[3], unitZ[3],
      r,phi,z; 

    if (cAlphaCoords == NULL) 
      {
	assert (relcAlphaCoords != NULL);
	numCAlphas     = lastAAIndexOfLastNonLoopVector_plusOne - 
	  firstAAIndexOfFirstNonLoopVector;
	cout << "numCalphas = " << numCAlphas << endl;
	cAlphaCoords   = new pFloat[numCAlphas];

	for (i = 0; i< numCAlphas; i++) 
	  {
	    cAlphaCoords[i] = new float[3];
	  }
  
	currentSseIndex =  0; 
	relAAIndex      = -1;
	findBaseAndUnitVecsForTheWholeSse(base, unitX, unitY, unitZ, 
					  this, currentSseIndex);
	
	cout << "base at : " << base[0] << " " << base[1] << " ";
	cout << base[2] << endl;
	cout << "unitX : " << unitX[0] << " " << unitX[1] << " " ;
	cout << unitX[2] << endl;
	cout << "unitY : " << unitY[0] << " " << unitY[1] << " " ;
	cout << unitY[2] << endl;
	
	for (i = 0; i < numCAlphas; i++) 
	  {
	    relAAIndex ++; 
	    if (relAAIndex == 
		((this->getSseElement(currentSseIndex))->getNumAAs())) 
	      {
		relAAIndex = 0;
		currentSseIndex ++;
		findBaseAndUnitVecsForTheWholeSse(base, unitX, unitY, unitZ,
						  this, currentSseIndex);

		cout << "base at : " << base[0] << " " << base[1] << " ";
		cout << base[2] << endl;
		cout << "unitX   : " << unitX[0] << " " << unitX[1] << " " ;
		cout << unitX[2] << endl;
		cout << "unitY : " << unitY[0] << " " << unitY[1] << " " ;
		cout << unitY[2] << endl;
	      }
	    
	    r   =  relcAlphaCoords[ i + firstAAIndexOfFirstNonLoopVector ][0];
	    phi =  relcAlphaCoords[ i + firstAAIndexOfFirstNonLoopVector ][1]
	      + 
	      ((this->getSseElement(currentSseIndex)) -> getTwistAngle());
	    z   =  relcAlphaCoords[ i + firstAAIndexOfFirstNonLoopVector ][2];
	    
	    for (j = 0; j< 3; j++) 
	      {
		cAlphaCoords[i][j] = base[j] + 
		  r * cos(phi) * unitX[j] + r * sin(phi) * unitY[j] 
		  + z * unitZ[j];
		/* 
		   or I could obtain inv(T) from unitX,Y,Z and base, and then 
		   compute inv(T) * [1 x1 y1 z1]', 
		   where x1 = r*cos(phi), z1 = z, etc.
		*/
	      }
	  }
      }
    else
      {
	cout << "CA coords not null. nothing to generate. " << endl;
      }
}
 
/* 
   similar to generateC-alphaCoords, except that this function is to be 
   called for every node. The entries corresponding to loops are set to -1.
   centroidCoords[0] corresponds to the AA that is in the first sse.
*/
 
void Protein::generateCentroidCoords() 
   {
     
  int i, j, numCentroids, 
    currentSseIndex,        // 0 .. # sseElements - 1
    relAAIndex ;            // 0.. numAAs in currentSseElement - 1
  
  float base[3],            // base of the vector corresponding to AAindex
    unitX[3],               /*
			      unit vec. corresponding to each amino acid 
			      in the current sse.
			    */
    unitY[3], unitZ[3],
    r,phi,z; // can remove these variables later
  
  
  assert(centroidCoords == NULL);
  
  numCentroids   = lastAAIndexOfLastNonLoopVector_plusOne - 
    firstAAIndexOfFirstNonLoopVector;
  
  centroidCoords = new pFloat[numCentroids];
  
  for (i = 0; i< numCentroids; i++) 
    {
      centroidCoords[i] = new float[3];
    }
  
  currentSseIndex = 0; relAAIndex = -1;
  findBaseAndUnitVecsForTheWholeSse(base, unitX, unitY, unitZ, 
				    this, currentSseIndex);
  for (i = 0; i< numCentroids; i++) 
    {
      
      relAAIndex ++;
      if (relAAIndex == 
	  ((this->getSseElement(currentSseIndex))->getNumAAs())) 
	{
	  relAAIndex = 0;
	  currentSseIndex ++;
	  findBaseAndUnitVecsForTheWholeSse(base, unitX, unitY, unitZ, this, currentSseIndex);
	}
      
      if (!loop[i + firstAAIndexOfFirstNonLoopVector]) 
	{
	  r   =  relCentroidCoords[i + firstAAIndexOfFirstNonLoopVector][0];
	  phi =  relCentroidCoords[i + firstAAIndexOfFirstNonLoopVector][1] + 
	    ((this->getSseElement(currentSseIndex)) -> getTwistAngle());
	  z   =  relCentroidCoords[i + firstAAIndexOfFirstNonLoopVector][2];
	  
	  for (j = 0; j< 3; j++) 
	    {
	      
	      centroidCoords[i][j] = base[j] +  r * cos(phi) * unitX[j] + r * sin(phi) * unitY[j] + z * unitZ[j];
	      /* or I could obtain inv(T) from unitX,Y,Z and base, and then 
		 compute inv(T) * [1 x1 y1 z1]', where x1 = r*cos(phi), z1 = z, etc.
	      */
	    }
	}
      else {
	assert( (i != 0) && (i != numCentroids - 1) );
	centroidCoords[i][0] = centroidCoords[i][1] = centroidCoords[i][2] = -1.0;
	//      assert(relCentroidCoords[i][0] == -1.0);
	//      assert(relCentroidCoords[i][1] == -1.0);
	//      assert(relCentroidCoords[i][2] == -1.0);
      }
    }
  
}
#if 0
  // for test purposes only, and for 1ROP
  void Protein::printFirstHelixCentroidCoordinates() {
    cout << "printing first helix centroid coordinates .. " << endl;
    for (int i = 2; i< 28; i++) {
      cout << centroidCoords[i][0] << " " << centroidCoords[i][1] << " " << centroidCoords[i][2] << endl;
    }
  }
#endif

/*
  u1 and u2 are 2 vectors 
*/

float findVectorAngle(Point A, Point B, Point C) 
{
  float vl1, vl2;
  
  Point u1,u2;
  
  VecSub(u1, A, B); VecSub(u2, C,  B); 
  
  vl1 = VecLength(u1); vl2 = VecLength(u2);
  assert((vl1 != 0.0) && (vl2 != 0.0));
  return (acos(VecDot(u1,u2) / vl1 / vl2));
}

#if 0
/*   
     computes bond and dihedral angles for the native structure. 
*/
 
void Protein::computeBondDihedralAngles() {
  int i;
  float vectorAngle, dihedralAngle;
  
  Point A,B,C,D;
  
  assert(numElements >= 3);

  VecCopy(A, sseVectors[0]->getBasePoint());
  VecCopy(B, sseVectors[0]->getEndPoint());
  VecCopy(C, sseVectors[1]->getEndPoint());
  VecCopy(D, sseVectors[2]->getEndPoint());
  
  vectorAngle  = findVectorAngle(A,B,C);
  sseVectors[1]->setVectorAngle(vectorAngle);
  
  for (i = 2 ; i< numElements - 1; i++) 
    {
      vectorAngle     = findVectorAngle    (B,C,D);
      dihedralAngle   = dihedral_angle  (A,B,C,D);
      sseVectors[i]->setVectorAngle    (vectorAngle);
      sseVectors[i]->setDihedralAngle(dihedralAngle);
      VecCopy(A,B);
      VecCopy(B,C);
      VecCopy(C,D);
      VecCopy(D, sseVectors[i+1]->getEndPoint());
    }
  
  vectorAngle     = findVectorAngle    (B,C,D);
  dihedralAngle   = dihedral_angle (A,B,C,D);
  sseVectors[i]->setVectorAngle        (vectorAngle);
  sseVectors[i]->setDihedralAngle    (dihedralAngle);
  if (VERBOSE_MODE)
    cout << "Native structure bond and dihedral angle computation is finished. " << endl;

}
#endif

/*   
     computes bond and dihedral angles for the native structure. 
*/
 
void Protein::computeBondDihedralAngles() 
{
  int i;
  float vectorAngle, dihedralAngle;
  
  Point A,B,C,D;
  
  assert (numElements >= 3);
  
  VecCopy(A, sseVectors[0]->getBasePoint());
  VecCopy(B, sseVectors[0]->getEndPoint());
  VecCopy(C, sseVectors[1]->getEndPoint());
  VecCopy(D, sseVectors[2]->getEndPoint());
  
  vectorAngle  = findVectorAngle(A,B,C);
  sseVectors[1]->setVectorAngle(vectorAngle);
  
  for (i = 2 ; i< numElements - 1; i++) 
    {
      vectorAngle     = findVectorAngle    (B,C,D);
      dihedralAngle   = dihedral_angle   (A,B,C,D);
      sseVectors[i]->setVectorAngle     (vectorAngle);
      sseVectors[i]->setDihedralAngle (dihedralAngle);
      VecCopy(A,B);
      VecCopy(B,C);
      VecCopy(C,D);
      VecCopy(D, sseVectors[i+1]->getEndPoint());
    }
  
  vectorAngle     = findVectorAngle    (B,C,D);
  dihedralAngle   = dihedral_angle   (A,B,C,D);
  sseVectors[i]->setVectorAngle        (vectorAngle);
  sseVectors[i]->setDihedralAngle    (dihedralAngle);
  
}

/* for debugging. Protein should aready have computed bond and 
   dihedral angles. This function goes over them and 
   computes from the vector locations the same angles and looks
   for the difference. */
void Protein::checkBondDihedralAngles() 
{
  int i;
  float vectorAngle, dihedralAngle;
  const float epsilon = 5E-3;

  Point A,B,C,D;

  assert(numElements >= 3);

  VecCopy(A, sseVectors[0]->getBasePoint());
  VecCopy(B, sseVectors[0]->getEndPoint());
  VecCopy(C, sseVectors[1]->getEndPoint());
  VecCopy(D, sseVectors[2]->getEndPoint());

  vectorAngle  = findVectorAngle(A,B,C);
  if (fabs(vectorAngle - sseVectors[1]->getVectorAngle()) > epsilon) {
    cout << "computed Bond     Angle = " << vectorAngle     << " precomputed value : " << sseVectors[1]->getVectorAngle() << endl;
    cout << "differs in first bondangle " << endl;
    cout << (*this);
    //uncomment after generating energy vs rmsd exit(0);

  }
  //  else cout << "bond angle 1 OK. " << endl;
  
  for (i = 2 ; i< numElements - 1; i++) {
    vectorAngle     = findVectorAngle    (B,C,D);
    dihedralAngle = dihedral_angle   (A,B,C,D);
    if (fabs(vectorAngle - sseVectors[i]->getVectorAngle()) > epsilon) {
      cout << "computed Bond Angle = " << vectorAngle << " precomputed value : " << sseVectors[i]->getVectorAngle() << endl;
      cout << "differs in " << i << " th bondangle" << endl;
      cout << (*this);
      //uncomment after generating energy vs rmsd exit(0);
    }
    
    if (fabs(dihedralAngle - sseVectors[i]->getDihedralAngle()) > epsilon) {
      cout << "computed Dihedral Angle = " << dihedralAngle << " precomputed value : " << sseVectors[i]->getDihedralAngle() << endl;
      cout << "Differ in " << i << " th element. " << endl;
      //uncomment after generating energy vs rmsd exit(0);
    }
  
    VecCopy(A,B);
    VecCopy(B,C);
    VecCopy(C,D);
    VecCopy(D, sseVectors[i+1]->getEndPoint());
  }

  vectorAngle     = findVectorAngle    (B,C,D);
  dihedralAngle = dihedral_angle (A,B,C,D);
  if (fabs(sseVectors[i]->getVectorAngle() - vectorAngle) > epsilon) {
    cout <<  "computed Bond Angle = " << vectorAngle << " precomputed value : " << sseVectors[i]->getVectorAngle() << endl;
    cout << "differs in last element. " << endl;
    cout << (*this);
    //    exit(0);
  }
  //  else    cout << "bond angle " << i << " OK. " << endl;

  if (fabs(dihedralAngle - sseVectors[i]->getDihedralAngle()) > epsilon) {
    cout << "computed Dihedral Angle = " << dihedralAngle << " precomputed value : " << sseVectors[i]->getDihedralAngle() << endl;
    cout << "Differ in last element. " << endl;
    //    exit(0);
  }
  //  else cout << "dihedralangle " << i << " OK. " << endl;

}

/* 
   called by standardizeNativeStructure, this function
   puts the first vectors basis at the origin, its end point at the 
   x axis, the second vector in the x-y plane, and recomputes the 
   end point of each following vector accordingly.
*/

void recomputeSseVectors (Protein * p)
{
  const Point origin = {0.0, 0.0, 0.0};
  Point endPoint;
  float length, vectorAngle, dihedralAngle;

  /* first Sse */
  
  endPoint[0] = endPoint[1] = 0.0; 
  endPoint[2] = p->getSseElement(0)->getLength();
  p->getSseElement(0)->setVector2(origin, endPoint);

  /* second Sse */
  
  length        = p->getSseElement(1)->getLength();
  vectorAngle   = p->getSseElement(1)->getVectorAngle();

  endPoint[0]   = 0.0;
  endPoint[2]   = p->getSseElement(0)->getEndPoint()[2] - 
    length * cos(vectorAngle);
  endPoint[1]   = sqrt(length * length - 
		       (endPoint[2] -  p->getSseElement(0)->getEndPoint()[2]) * 
		       (endPoint[2] -  p->getSseElement(0)->getEndPoint()[2]));
  p->getSseElement(1)->setVector2(p->getSseElement(0)->getEndPoint(), 
				  endPoint);
  
  /* other Sses */
  for (int i = 2; i < p->getNumSseElements(); i++) 
    {
      dihedralAngle   = p->getSseElement(i)->getDihedralAngle();
      vectorAngle     = p->getSseElement(i)->getVectorAngle();
      length          = p->getSseElement(i)->getLength();
      
      obtainFourthPoint(p->getSseElement(i-2)->getBasePoint(), 
			p->getSseElement(i-2)->getEndPoint(), 
			p->getSseElement(i-1)->getEndPoint(), endPoint, 
			dihedralAngle, vectorAngle, length);
      
      p->getSseElement(i)->setVector2(p->getSseElement(i-1)->getEndPoint(), 
				      endPoint);
    }
}

/* 
   first copies the config. (i.e. angles, lengths) 
   of the original protein to the new protein.
   
   then recomputes the vectors to put the first vector basis
   at the origin, etc.
*/
void standardizeNativeStructure(Protein * standardP, 
				Protein * originalP) 
{
  standardP -> copyConfig  (originalP);
  recomputeSseVectors      (standardP);
}

#if 0
  void Protein::createTotallyElongatedConfig(const Protein * myProtein) {
    int i;
    const Point originPoint = {0.0, 0.0, 0.0};
    const float  vectorAngle = M_PI       , dihedralAngle = M_PI   , twistAngle = 0.0;
    Point  endPoint, basePoint;

    elements      = new pSse[numElements];
  

    for (i = 0; i< numElements; i++) {
   
      elements[i] = new Sse (myProtein->elements[i]->getType(), originPoint, originPoint, myProtein->elements[i]->getNumAAs());
    
      elements[i]->setVectorAngle (vectorAngle);
      elements[i]->setTwistAngle(twistAngle);
      elements[i]->setDihedralAngle(dihedralAngle);

      switch(i) {
      case 0 : {
	endPoint[0] = endPoint[1] = 0.0; endPoint[2] = myProtein->elements[0]->getLength();
	//  cout << "end pt. of sse element at : " << endPoint[2] << endl;
	elements[0]->setVector2(originPoint, endPoint);
	break;
      }
      default: {
	VecCopy(basePoint, elements[i-1]->getEndPoint());
	endPoint[0] = elements[i-1]->getEndPoint()[0];
	endPoint[1] = elements[i-1]->getEndPoint()[1];
	endPoint[2] = elements[i-1]->getEndPoint()[2] + myProtein->elements[i]->getLength();
	elements[i]->setVector2(basePoint, endPoint);
      }
      } // switch
    }
    computeEnergy();
    //  setRMSDField(allNodes[0]);

  }
#endif

  /* allocates space, then reads the files one by one */
void  Protein::readRelativeCprime_N_OPositions(const char CprimeFilename[], 
					       const char NFilename[],
					       const char OFilename[]) 
{
  int i;
  int sequenceSize;
  
  sequenceSize = sequence.size();
  
  relCprimeCoords  = new pFloat[sequenceSize];
  relNCoords       = new pFloat[sequenceSize];
  relOCoords       = new pFloat[sequenceSize];
  
  for (i = 0; i< sequenceSize; i++) 
    {
      relCprimeCoords[i] = new float[3];
      relNCoords[i]      = new float[3];
      relOCoords[i]      = new float[3];
    }
    
  /* reading Cprime */
  ifstream fin(CprimeFilename);
  if (!fin) 
    {
      cout << CprimeFilename << " was not opened. " << endl;
      exit(0);
    }
  else
    {
      cout << "reading " << CprimeFilename << endl;
    }

  i = 0;
  while ((!fin.eof()) && (i<sequenceSize)) 
    {
      fin >> relCprimeCoords[i][0] >> relCprimeCoords[i][1] >> relCprimeCoords[i][2];
      i++;
      if (fin.eof())
	break;
    }
  assert(i == sequenceSize);
  fin.close();

  /*  reading N */

  fin.open(NFilename);
    
  if (!fin) 
    {
      cout << NFilename << " was not opened. " << endl;
      exit(0);
    }
  else
    {
      cout << "reading " << NFilename << endl;
    }

  i = 0;
  while ((!fin.eof()) && (i<sequenceSize)) 
    {
      fin >> relNCoords[i][0] >> relNCoords[i][1] >> relNCoords[i][2];
      i++;
      if (fin.eof())
	break;
    }
  assert(i == sequenceSize);

  fin.close();

  /* reading O */
  fin.open(OFilename);
    
  if (!fin) 
    {
      cout << OFilename << " was not opened. " << endl;
      exit(0);
    }
  else
    {
      cout << "reading " << OFilename << endl;
    }

  i = 0;
  while ((!fin.eof()) && (i<sequenceSize)) 
    {
      fin >> relOCoords[i][0] >> relOCoords[i][1] >> relOCoords[i][2];
      i++;
      if (fin.eof())
	break;
    }
    
  assert(i == sequenceSize);
  fin.close();

}

// C', N and O coordinates are sequentially generated. 
void Protein::generateCprime_N_OPositions() 
{
  int i, j, k, numPositions, 
    currentSseIndex, // 0 .. # sseElements - 1
    relAAIndex ;            // 0.. numAAs in currentSseElement - 1
  float base[3],            // base of the vector corresponding to AAindex
    unitX[3],                /*
			       unit vec. corresponding to
			       each amino acid in the current sse.
			     */
    unitY[3], unitZ[3],
    r,phi,z; 
  
  if (NCoords == NULL) 
    {
      numPositions = lastAAIndexOfLastNonLoopVector_plusOne - firstAAIndexOfFirstNonLoopVector;
      NCoords = new pFloat[numPositions];
      for (i = 0; i< numPositions; i++) 
	{
	  NCoords[i] = new float[3];
	}
      
      assert(OCoords == NULL);
      OCoords = new pFloat[numPositions];
      for (i = 0; i< numPositions; i++) 
	{
	  OCoords[i] = new float[3];
	}
      
      assert(CPrimeCoords == NULL);
      CPrimeCoords = new pFloat[numPositions];
      for (i = 0; i< numPositions; i++) 
	{
	  CPrimeCoords[i] = new float[3];
	}
      
      for (k = 0; k < 3; k++) 
	{
	  /* 
	     k == 0 -> c'
	     k == 1 -> N
	     k == 2 -> O coords are sought. 
	  */
	  
	  currentSseIndex = 0; relAAIndex = -1;
	  findBaseAndUnitVecsForTheWholeSse(base, unitX, unitY, unitZ, 
					    this, currentSseIndex);
	
	  for (i = 0; i< numPositions; i++) 
	    {
	      relAAIndex ++; 
	      if (relAAIndex == 
		  ((this->getSseElement(currentSseIndex))->getNumAAs())) 
		{
		  relAAIndex = 0;
		  currentSseIndex ++;
		  findBaseAndUnitVecsForTheWholeSse(base, unitX, unitY, unitZ, 
						    this, 
						    currentSseIndex);
		}
  
	      switch(k) 
		{
		case 0 : {
		  r   =  relCprimeCoords[i +  firstAAIndexOfFirstNonLoopVector][0];
		  phi =  relCprimeCoords[i +  firstAAIndexOfFirstNonLoopVector][1] + 
		    ((this->getSseElement(currentSseIndex)) -> getTwistAngle());
		  z   =  relCprimeCoords[i +  firstAAIndexOfFirstNonLoopVector][2];
	    
		  for (j = 0; j< 3; j++) {
	      
		    CPrimeCoords[i][j] = base[j] + 
		      r * cos(phi) * unitX[j] + r * sin(phi) * unitY[j] + z * unitZ[j];
		  }
	    
		  break;
		}
		case 1 : {
		  r   =  relNCoords[i +  firstAAIndexOfFirstNonLoopVector][0];
		  phi =  relNCoords[i +  firstAAIndexOfFirstNonLoopVector][1] + 
		    ((this->getSseElement(currentSseIndex)) -> getTwistAngle());
		  z   =  relNCoords[i +  firstAAIndexOfFirstNonLoopVector][2];
	    
		  for (j = 0; j< 3; j++) {
	      
		    NCoords[i][j] = base[j] + 
		      r * cos(phi) * unitX[j] + r * sin(phi) * unitY[j] + z * unitZ[j];
		  }
	    
		  break;
		}
		case 2 : { 
		  r   =  relOCoords[i +  firstAAIndexOfFirstNonLoopVector][0];
		  phi =  relOCoords[i +  firstAAIndexOfFirstNonLoopVector][1] + 
		    ((this->getSseElement(currentSseIndex)) -> getTwistAngle());
		  z   =  relOCoords[i +  firstAAIndexOfFirstNonLoopVector][2];
	    
		  for (j = 0; j< 3; j++) {
	      
		    OCoords[i][j] = base[j] + 
		      r * cos(phi) * unitX[j] + r * sin(phi) * unitY[j] + z * unitZ[j];
		  }
		  break;
		}
		}//switch
	    }
	}
    }
}

  void Protein::reportEnergy() {
    cout << "H-H energy = " << HHEnergy;
    cout << " electrostatic Energy : " << coulombicEnergy << endl;
    cout << " due to individual charges: " << indivChargesCoulombicContribution;
    cout << " due to alpha helix : " << alphaHelixCoulombicContribution << endl;
    cout << " exclusion Energy = " << exclusionEnergy;
    cout << "hydrogen bond energy : " <<  hBondEnergy << endl;
    cout << " total : " << energy << endl;

  }

  int Protein::getIndex() const {

    return index;
  }

  /* only prints parameters, and to stdout. */
  void Protein::printParameters() const
    {

      int i;
      int numSseElements = getNumSseElements();
      Sse * mySse;


      cout << "printing parameters of the protein... " << endl;

      for (i = 0; i < numSseElements; i++) 
	{
	  mySse = getSseElement(i);
	  cout << mySse -> getDihedralAngle() << "," << 
	    mySse->getVectorAngle () << "," << 
	    mySse->getTwistAngle() << "," << 
	    mySse->getLength() << ",";
	}
      cout << endl;
    }

  /* 
     fout is already opened, and will not be closed at the end of operation.
     prints parameters of the protein, as well as its energy and rmsd from native. 
  */

  void Protein::printParameters(FILE * fout) const {

    int i;
  
    for (i = 0; i < getNumSseElements(); i++) {
      fprintf(fout, "%f , %f , %f , %f ,", 
	      getSseElement(i)->getDihedralAngle(),
	      getSseElement(i)->getVectorAngle (),
	      getSseElement(i)->getTwistAngle(),
	      getSseElement(i)->getLength());
    }
  
    fprintf(fout, "\n");
    //  fprintf(fout, "Energy = %f RMSD from native : %f \n", energy, rmsdFromNative); 

  }

  void Protein::printParameters(const string& nodeFilename) const 
    {

      FILE * fout;
  
      fout = fopen(nodeFilename.c_str(), "a");
      if (fout == NULL)
	{
	  fprintf(stderr, "%s could not be opened.\n", nodeFilename.c_str());
	  exit(0);
	}
  
      fprintf(fout, "node : %d\n", getIndex());
      for (int i = 0; i < numElements; i++) 
	{
	  fprintf(fout, "%f %f %f %f\n",  sseVectors[i]->getDihedralAngle(),
		  sseVectors[i]->getVectorAngle (),
		  sseVectors[i]->getTwistAngle(),
		  sseVectors[i]->getLength());
	}
      fprintf(fout, "%f\n", getEnergy());
      fclose(fout);
  
    }
#if 0
  void Protein::setRMSDField(const Protein * nativeP) {
    rmsdFromNative = this->computeRMSD(nativeP);

  }
#endif




  /* calls C-alpha, C', N, O  coordinates calculation routines. */
  void   Protein::generateAtomCoords() {
    generateCAlphaCoords();
    generateCprime_N_OPositions();

  }

  /* prints non-loop C-alpha, C' , N O coordinates to file. */
  void   Protein::printAtomCoords (char * filename) 
    {
      int  numAtomsToWrite,i;
      FILE * fout;

      fout  = fopen(filename, "w");
  
      cout << "protein coords written to : " << filename << endl;
  
      numAtomsToWrite =  Protein::lastAAIndexOfLastNonLoopVector_plusOne- 
	Protein::firstAAIndexOfFirstNonLoopVector;
      
      //  fprintf(fout, "protein energy = %f rmsd from native : %f\n", energy, rmsdFromNative);
  
      for (i = 0; i< numAtomsToWrite; i++) 
	{

	  if (!(loop[i + firstAAIndexOfFirstNonLoopVector])) 
	    {
      
	      fprintf(fout, "CA %f %f %f \n", this->cAlphaCoords[i][0],this->cAlphaCoords[i][1],this->cAlphaCoords[i][2]);
	      fprintf(fout, "C' %f %f %f \n", this->CPrimeCoords[i][0],this->CPrimeCoords[i][1],this->CPrimeCoords[i][2]);
	      fprintf(fout, "N %f %f %f \n",  this->NCoords[i][0],     this->NCoords[i][1],this->NCoords[i][2]);
	      fprintf(fout, "O %f %f %f \n",  this->OCoords[i][0],     this->OCoords[i][1],this->OCoords[i][2]);
	    }
    
	}
      fclose(fout);

    }
#if 0
  // relAAindex = 0.. numAAs in current SSe element - 1
  void findBaseAndUnitVecs(float base[], float unit[], Protein *myProtein, int currentSseIndex, int relAAIndex ) {
  
    float temp[3];

    assert((relAAIndex >=0) && (relAAIndex < (myProtein->getSseElement(currentSseIndex)) -> getNumAAs()));
    VecCopy(base, (myProtein->getSseElement(currentSseIndex))->getBasePoint());
    VecSub (unit, (myProtein->getSseElement(currentSseIndex))->getEndPoint(), (myProtein->getSseElement(currentSseIndex))->getBasePoint());
    assert((myProtein->getSseElement(currentSseIndex)->getNumAAs()) != 0);
    VecScale(unit, (1.0/(myProtein->getSseElement(currentSseIndex)->getNumAAs())), unit);
    VecScale(temp, relAAIndex, unit);
    VecAdd(base, base, temp);
    /* base and unit vector for the amino acid are found. */
  }
#endif

#if 0
  void Protein::saveAPoint(int * numAtoms, Point * points, const char * line) {

    char record[10];

    extract(points[*numAtoms], line);
    /*

      extStr(record,line,119,122);
      points[*numAtoms] [0] = (float) (atof(record));
      extStr(record,line,126,129);
      points[*numAtoms] [1] = (float) (atof(record));
      extStr(record,line,133,136);
      points[*numAtoms] [2] = (float) (atof(record));
    */
    cout << "Saved: (" << points[*numAtoms] [0] << "," ;
    cout << points[*numAtoms] [1] << "," ;
    cout << points[*numAtoms] [2] << ") . Atom #" << *numAtoms << endl ;
    (*numAtoms) ++;

  }

#endif
#if 0
  void   setHPMapping() {

    const char hydrophobics[] = {'A','V', 'F', 'P', 'M', 'I', 'L', 'G'};
    const char polars[]       = {'D','E','K','R', 'S','T','Y','H','C','N','Q','W'};
    vector<char> H(hydrophobics, &hydrophobics[sizeof(hydrophobics)]);
    vector<char> P(polars, &polars[sizeof(polars)]);

    for (int i = 0; i< H.size(); i++) {
      cout << H[i];
    }
    cout << endl;
    for (i = 0; i< P.size(); i++) {
      cout << P[i] ;
    }
    cout << endl;

    for (i = 0; i< H.size(); i++) {
      Protein::HP_mapping[H[i]] = true;
    }
    for (i = 0; i< P.size(); i++) {
      Protein::HP_mapping[P[i]] = false;
    }
  
  }
#endif
#if 0



  void  Protein::allocateGridCoordinatesVector() {
    assert(gridCoordinates.size() == 0);
    gridCoordinates.resize(numElements);
    for (int i = 0; i< numElements; i++) {
      gridCoordinates[i].resize(sseVectors[i]->getNumAAs());
      for (int j = 0; j< gridCoordinates[i].size(); j++) {
	gridCoordinates[i][j].resize(3);
      }
    }
  }

#endif
  inline void Protein::testInvariant() const throw()
    {
      assert(index >= -1);
      for (unsigned int i = 0; i < params.size(); i++)
	{
	  assert((params[i] >= 0) && (params[i] <= 1));
	}
      assert(params.size() == ANN_dim);
  
      /* index can be -1 */
      //  assert ((maxEnergy >= -30) && (maxEnergy <= 200));
      //  assert(exclusionEnergy <= maxExEnergy); /* reasonable limits? */
      //  assert(rmsdFromNative >= 0);
    }
 
  float Protein::computeDistance(const Protein * otherProtein) const
    {
      float retval = 0.0;
      for (unsigned int i = 0; i < params.size(); i++)
	{
	  float temp = params[i] - otherProtein->params[i];
	  temp      *= temp;
	  retval    += temp;
	}
      return sqrt(retval);
    }


  /* normalizes the value between min and max to between 0 and 1.
     returned value is (value - min)/(max - min)
  */
  inline float normalized(float value, float min, float max)
    {
      assert(max -  min != 0.0);
      float retval = (value - min) / (max - min);
      return retval;
    }

  /* 
     given dihedral, twist, and bond angles, and also the vector length, 
     sets the params vector of protein which holds normalized parameters of
     the protein.
  */
void Protein::findParams ()
{
  
  this->params.resize(0);
  int numElements = this->getNumSseElements();
  
  for (int j = 0; j <  numElements; j++) 
    {
      Sse * sse_j = this->getSseElement(j);
      float dih   = sse_j->getDihedralAngle();
      float twi   = sse_j->getTwistAngle();
      float veA   = sse_j->getVectorAngle();
      float len   = sse_j->getLength();
      
      if (sse_j->getType() == LOOP)
	{
	  int numAAs = sse_j->getNumAAs();
	  this->params.push_back( normalized(len, 
					     numAAs * 
					     Protein::minLengthPerAA,
					     numAAs * 
					     Protein::maxLengthPerAA));
	}
      else // ALPHA OR BETA
	{
	  assert((sse_j->getType() == ALPHA) || 
		 (sse_j->getType() == BETA));
	  /* 
	     twist angle is a DOF valid for every vector,
	     including the first one. 
	  */
	  this->params.push_back(normalized(twi,  -M_PI, M_PI));
	}
      
      switch(j)
	{
	case 0 :
	  {
	    /* no DOF to store. */
	    break;
	  }
	case 1 :
	  {
	    /* store only vector angle */
	    this->params.push_back(normalized(veA, 0.0, M_PI));
	    break;
	  }
	default : 
	  {
	    /* store both vector and dihedral angles */
	    assert(j >= 2);
	    this->params.push_back( normalized(veA, 0.0, M_PI));
	    this->params.push_back( normalized(dih, -M_PI, M_PI));
	  }
	}
    }
  assert(this->params.size() ==  Protein::NUMDOF);
}




const vector<float>& Protein::getParams() const
{
  return params;
}

/*
    sets pDSqr. pDSqr holds the square of the distance between the residues
    which are on the same
    alpha helix and also beta strands.
    pD[i][j] = distance; means, the C-alpha distance between 
    ith and jth residues is dist. i and j should be residues on 
    the same alpha helix or beta strand vector. Otherwise, distance
    is -1. Distance is also -1 for AAs not on alpha helices or beta strands.
*/

void Protein::setPairwiseDistances()
{
  int i, j;
  static int numCalls = 0;
  
  /* 
     this function should only be called once. 
  */
  
  assert (numCalls == 0);
  numCalls ++;
  
  cout << "setting up the data structure to precompute pairwise centroid distances on the same SSE." << endl;
  
  int numAminoAcids = getNumAAs();
  
  pDSqr.resize(numAminoAcids);
  
  for (i = 0; i < numAminoAcids; i++)
    {
      pDSqr[i].resize (numAminoAcids);
    }
  
  for (i = 0; i < numAminoAcids; i++)
    {
      if (!loop [i])
	{
	  for (j = i; j < numAminoAcids; j++)
	    {
	      if (!loop  [j])
		{
		  float distSqr = 
		    distanceSquaredBetweenCentroids (
						     i - firstAAIndexOfFirstNonLoopVector , 
						     j - firstAAIndexOfFirstNonLoopVector );
		  pDSqr[i][j] = pDSqr [j] [i] =  distSqr;
		}
	      else
		// jth AA is loop, ith is not, 
		// connection of jth AA with ith will be set to -1.
		// since for each i not part of loop. jth AA will be 
		// checked, no need to set all connections of jth AA to -1.
		{
		  pDSqr[i][j] = pDSqr [j] [i] = -1.0;
		}
	    }
	}
      else // ith AA is loop, all its distances will be set to -1.
	{
	  for (j = 0; j < numAminoAcids; j++)
	    {
	      pDSqr[i][j] = pDSqr [j] [i] = -1.0;
	    }
	}
    }
  
  keepDistancesWithinTheSameSSE (pDSqr);
}

/* 
   sets all distances not within the same SSE to -1. 
*/

void keepDistancesWithinTheSameSSE (vector<vector<float> > &pDSqr)
{
  int i = 0;
  
  int numAminoAcids = Protein::getNumAAs();

  while (i < numAminoAcids)
    {
      if (!Protein::loop[i]) // a sse started.
	{
	  int j = i;
	  
	  while ((!Protein::loop[j]) && (j < numAminoAcids))
	    j++;

	  // sse starts from i, extends till j-1.
	  // so delete all connections with i..j-1 except
	  // the ones [a,b] where a and b E [i..j-1].
	  
	  //cout << "SSE from " << i << " th AA till " << j-1 << " th AA." << endl;
	  assert (j - 1 >= i);
	  deleteAllConnectionsExceptFrom (i, j-1, pDSqr);
	  i = j - 1;
	  /*
	    will increase i by 1 so will restore 
	    from non-loop residue.
	  */
	}
      else 
	{
	  /*
	    ith AA is a loop AA. So delete all its distances 
	    with all other residues.
	  */
	  //	  cout  << i << " th AA is on a loop." << endl;
	  deleteAllConnections (i, pDSqr);
	}
      i++;
    }
#if 0
  cout << "for debugging, printing the pD array : " << endl;
  for (i = 0; i < numAminoAcids; i ++)
    {
      cout << "i = " << i << " : ";
      for (int j = 0; j < numAminoAcids; j ++)
	{
	  cout << pDSqr [i] [j] << " ";
	}
      cout << endl;
    }
#endif
}

  /*
    set all pairwise distances involving i'th AA to -1.
    So set pD[i][j] = -1.0. for all j, including i. 
    make sure pDSqr matrix is symmetric.
  */
  void deleteAllConnections (int i,  vector<vector<float> > & pDSqr)
    {
      int numAminoAcids = Protein::getNumAAs();
  
      for (int j = 0; j < numAminoAcids; j ++)
	{
	  pDSqr [j][i] = pDSqr [i] [j] = -1.0;
	}
    }
  /* 
     all connections from i..j  will be kept, and others will be set to -1.
     i.e., pDSqr [k][a] = -1.0, where a E [i..j], k E [0..i-1] U [j+1.. numAAs - 1].
     also will maintain the symmetry of the matrix.
  */

void deleteAllConnectionsExceptFrom (int i, int j, 
				     vector<vector<float> > & pDSqr)
{
  int a, k;
  
  int numAminoAcids = Protein::getNumAAs();
  
  assert ((i >= 0) && (i < numAminoAcids));
  assert ((j >= 0) && (j < numAminoAcids));
  
  for    (a = i; a <= j; a ++)
    {
      for (k = 0; k <= i-1; k++)
	{
	  pDSqr [k] [a] = pDSqr [a] [k] = -1.0; 
	}
      for (k = j + 1; k < numAminoAcids; k++)
	{
	  pDSqr [k] [a] = pDSqr [a] [k] = -1.0; 
	}
    }
}

  /* 
     this is the routine that does the real energy computation.
     computeEnergy calls this.
     segregated like this since we want to be able to compute the energy
     sometimes w/o generating centroid coordinates.
  */
float Protein::_computeEnergy() 
{
  int k,l, numCentroids2Cover;
  float temp;
  
  
  energy = 0.0;
  
  numCentroids2Cover = lastAAIndexOfLastNonLoopVector_plusOne- 
    firstAAIndexOfFirstNonLoopVector;
  
  HHEnergy    =  coulombicEnergy = exclusionEnergy = 0.0;
  hBondEnergy = 0.0;
  numComputations++;
  
  const int nLAIS = nonLoopAAIndices.size(); // non-loop aa indices vector size

  for (k = 0; k < nLAIS; k++) 
    {
      int nonLoopAAIndexK     = nonLoopAAIndices[k];
#if USE_HYDROPHOBIC_ENERGY
      bool hyd_k = hydrophobic[sequence[nonLoopAAIndexK]];
#endif
      
      assert     (!loop[nonLoopAAIndexK]);
      
      for        (l = k + 1; l < nLAIS ; l++) 
	{
	  int nonLoopAAIndexL = nonLoopAAIndices[l];
	  assert (!loop[nonLoopAAIndexL] );
	  // float cAlpha_distance_sqr = distanceSquaredBetweenCAlphas  (i,j);
	  float centroidDistanceSqr;
#if 0
	  /*
	    the below routine is for debugging the time consumed in
	    the two ways of computing the distance.. it seems 
	    lookup is slower!!!
	  */
	  struct tms start,end;
	  const int numIter = 1E7;
	  times      (&start);
	  for (int kk = 0; kk < numIter; kk++)
	    {
	      centroidDistanceSqr  = 
		distanceSquaredBetweenCentroids(
						nonLoopAAIndexK - firstAAIndexOfFirstNonLoopVector,
						nonLoopAAIndexL - firstAAIndexOfFirstNonLoopVector);
	    }
	  times       (&end);
	  cout << " time for " << numIter << " computation of the distance sqr :" << endl;
	  reportTimes (start, end);
	  //	  cout << "dist = " << centroidDistanceSqr << endl;

	  times      (&start);
	  for (int kk = 0; kk < numIter; kk++)
	    {
	      centroidDistanceSqr  = pDSqr [nonLoopAAIndexL]
		[nonLoopAAIndexK];
	    }
	  times      (&end);
	  cout << " time for " << numIter << " lookup :" << endl;
	  reportTimes (start, end);
	  //	  cout << "dist = " << centroidDistanceSqr << endl;
	  //	  exit(0);
	  /* end debug routine.. */
#endif
	  
	  float lookupDistance = 
	    pDSqr[nonLoopAAIndexL][nonLoopAAIndexK];

	  if (lookupDistance == -1.0)
	    {
	      centroidDistanceSqr  = 
		distanceSquaredBetweenCentroids(
						nonLoopAAIndexK - 
						firstAAIndexOfFirstNonLoopVector,
						nonLoopAAIndexL - 
						firstAAIndexOfFirstNonLoopVector);
	    }
	  else
	    {
	      // Use precomputed value...
	      centroidDistanceSqr  = lookupDistance;
	      /*
		float debugDist = 
		distanceSquaredBetweenCentroids(
		nonLoopAAIndexK - firstAAIndexOfFirstNonLoopVector,
		nonLoopAAIndexL - firstAAIndexOfFirstNonLoopVector);
		//  cout << "abs diff : " << fabs (debugDist - centroidDistanceSqr) << endl;
		assert (fabs (debugDist - centroidDistanceSqr) < 1E-3);
	      */
	    }
	  //energy           += exclusionEnergy_Calpha(cAlpha_distance_sqr);
	  temp                = exclusionEnergy_Centroid(centroidDistanceSqr);
	  energy             += temp;
	  exclusionEnergy    += temp;
	  if (exclusionEnergy > maxExEnergy) // break early if too many clash
	    {
	      return energy;
	    }
#if USE_HYDROPHOBIC_ENERGY
	  if ((hyd_k)
	      && 
	      (hydrophobic[sequence[nonLoopAAIndexL]])) 
	    {
	      temp       = HP_Factor_Altered(centroidDistanceSqr);
	      energy    += temp;
	      //energy += HP_Factor_Altered(cAlpha_distance_sqr);
	      HHEnergy  += temp;
	    }
#endif
	}
    }
#if INCLUDE_HYDROGEN_BONDS 
  hBondEnergy = computeHBondEnergyTerm();
  energy     += hBondEnergy;
#endif

  
#if 0
  
  for (i = 0;  i < getNumSseElements(); i++) {
    if (getSseElement(i)->getType() == ALPHA) {
      energy += coulombic(getSseElement(i)->getBasePoint(), 0.5, getSseElement(i)->getEndPoint(), -0.5 );
      for (j = i+1; j < getNumSseElements(); j++) {
	
	if (getSseElement(j)->getType() == ALPHA) {

	  energy += coulombic(getSseElement(i)->getBasePoint(),  0.5, 
			      getSseElement(j)->getBasePoint(),  0.5);
	  energy += coulombic(getSseElement(i)->getBasePoint(),  0.5, 
			      getSseElement(j)->getEndPoint(),  -0.5);
	  energy += coulombic(getSseElement(i)->getEndPoint() , -0.5, 
			      getSseElement(j)->getBasePoint(),  0.5);
	  energy += coulombic(getSseElement(i)->getEndPoint() , -0.5, 
			      getSseElement(j)->getEndPoint(),  -0.5);
	}
      }
    }
  }

  alphaHelixCoulombicContribution = energy - (exclusionEnergy + HHEnergy);
  
  for (i = 0; i < numCentroids2Cover; i++) {
    if (charged[i + firstAAIndexOfFirstNonLoopVector]) {
      for (j = i+1; j < numCentroids2Cover; j++) {
	if (charged[j + firstAAIndexOfFirstNonLoopVector]) {
	  energy += coulombic(centroidCoords[i], charge[i + firstAAIndexOfFirstNonLoopVector], 
			      centroidCoords[j], charge[j + firstAAIndexOfFirstNonLoopVector]);
	}
      }
    }
  }
  
  coulombicEnergy = energy - (exclusionEnergy + HHEnergy);
  indivChargesCoulombicContribution = coulombicEnergy - alphaHelixCoulombicContribution;
#endif
  return energy;
}

  /* 
     generate the coords of the H connected to N. 
     the generateion uses a heuristic, by computing the unit vector from C to O
     and adding it to N. 
  */
#if 0
void Protein::generateHCoords()
{
  int i;
  if (HCoords == NULL)
    {
      int numPositions = 
	lastAAIndexOfLastNonLoopVector_plusOne - 
	firstAAIndexOfFirstNonLoopVector;

      HCoords = new pFloat[numPositions];
      
      for (i = 0; i < numPositions; i++) 
	{
	  HCoords[i] = new float[3];
	}
      
      Point Cprime2O;
      for (i = 0; i < numPositions; i++) 
	{
	  VecSub       (Cprime2O, OCoords[i], CPrimeCoords[i]);
	  VecNormalize (Cprime2O);
	  assert (fabs(VecLength (Cprime2O) - 1.0) < 1E-3);
	  VecAdd       (HCoords[i], NCoords[i], Cprime2O);
	}
    }
}
#endif
#if 0
/* 
   only generates H-O-C'-N coords that are involved in H-bonds.
*/
void Protein::generateCprime_N_O_H_Coords_InvolvedInHBonds ()
{
  int i, j, k, numPositions, 
    currentSseIndex,        // 0 .. # sseElements - 1
    relAAIndex ;            // 0.. numAAs in currentSseElement - 1
  float base[3],            // base of the vector corresponding to AAindex
    unitX[3],               /*
			      unit vec. corresponding to
			      each amino acid in the current sse.
			    */
    unitY[3], unitZ[3],
    r,phi,z; 

  
#if 0
  /*  
      generate each relevant C'-O-N-H;
      depending on the type;
      keep them in an easily accessible array.
      then compute angle;
      compute distance;
      add -1 if ok.

      also: gaussian sampling.
      acceleration of centroid coord computation.
  */
#endif

  if (NCoords == NULL) 
    {
      numPositions = lastAAIndexOfLastNonLoopVector_plusOne - firstAAIndexOfFirstNonLoopVector;
      NCoords = new pFloat[numPositions];
      for (i = 0; i< numPositions; i++) 
	{
	  NCoords[i] = new float[3];
	}
      
      assert(OCoords == NULL);
      OCoords = new pFloat[numPositions];
      for (i = 0; i< numPositions; i++) 
	{
	  OCoords[i] = new float[3];
	}
      
      assert(CPrimeCoords == NULL);
      CPrimeCoords = new pFloat[numPositions];
      for (i = 0; i< numPositions; i++) 
	{
	  CPrimeCoords[i] = new float[3];
	}
      
      for (k = 0; k < 3; k++) 
	{
	  /* 
	     k == 0 -> c'
	     k == 1 -> N
	     k == 2 -> O coords are sought. 
	  */
	  
	  currentSseIndex = 0; relAAIndex = -1;
	  findBaseAndUnitVecsForTheWholeSse(base, unitX, unitY, unitZ, 
					    this, currentSseIndex);
	
	  for (i = 0; i< numPositions; i++) 
	    {
	      relAAIndex ++; 
	      if (relAAIndex == 
		  ((this->getSseElement(currentSseIndex))->getNumAAs())) 
		{
		  relAAIndex = 0;
		  currentSseIndex ++;
		  findBaseAndUnitVecsForTheWholeSse(base, unitX, unitY, unitZ, 
						    this, 
						    currentSseIndex);
		}
  
	      switch(k) 
		{
		case 0 : {
		  r   =  relCprimeCoords[i +  firstAAIndexOfFirstNonLoopVector][0];
		  phi =  relCprimeCoords[i +  firstAAIndexOfFirstNonLoopVector][1] + 
		    ((this->getSseElement(currentSseIndex)) -> getTwistAngle());
		  z   =  relCprimeCoords[i +  firstAAIndexOfFirstNonLoopVector][2];
	    
		  for (j = 0; j< 3; j++) {
	      
		    CPrimeCoords[i][j] = base[j] + 
		      r * cos(phi) * unitX[j] + r * sin(phi) * unitY[j] + z * unitZ[j];
		  }
	    
		  break;
		}
		case 1 : {
		  r   =  relNCoords[i +  firstAAIndexOfFirstNonLoopVector][0];
		  phi =  relNCoords[i +  firstAAIndexOfFirstNonLoopVector][1] + 
		    ((this->getSseElement(currentSseIndex)) -> getTwistAngle());
		  z   =  relNCoords[i +  firstAAIndexOfFirstNonLoopVector][2];
	    
		  for (j = 0; j< 3; j++) {
	      
		    NCoords[i][j] = base[j] + 
		      r * cos(phi) * unitX[j] + r * sin(phi) * unitY[j] + z * unitZ[j];
		  }
	    
		  break;
		}
		case 2 : { 
		  r   =  relOCoords[i +  firstAAIndexOfFirstNonLoopVector][0];
		  phi =  relOCoords[i +  firstAAIndexOfFirstNonLoopVector][1] + 
		    ((this->getSseElement(currentSseIndex)) -> getTwistAngle());
		  z   =  relOCoords[i +  firstAAIndexOfFirstNonLoopVector][2];
	    
		  for (j = 0; j< 3; j++) {
	      
		    OCoords[i][j] = base[j] + 
		      r * cos(phi) * unitX[j] + r * sin(phi) * unitY[j] + z * unitZ[j];
		  }
		  break;
		}
		}//switch
	    }
	}
    }
}
#endif
/*
  set up the HH bond list. the file is composed of lines, 
      of the  following format : 
    <type> <AA1> <AA2>.
    type is 1 or 2. 
    type is 1 for the case where the AA with index AAIndex has the 
    N-H and otherAAIndex has the O. 2 for the case where AA with index 
    AAIndex has the O and the other AA has the N-H.
*/
void Protein::setupHHBondsList(string filename)
{
  FILE * fin = fopen (filename.c_str(), "r");

  if (fin == NULL)
    {
      cout << "couldn't read the file : " << filename << endl;
      exit(0);
    }
  
  while (!feof(fin))
    {
      int type, AA1, AA2;

      fscanf (fin, "%d %d %d\n", &type, &AA1, &AA2);

      HHBondType hhbond;
  
      hhbond.type = type;
      hhbond.AA1  = AA1;
      hhbond.AA2  = AA2;

      HHBondsList.push_back (hhbond);
    }
  
  cout << "read " <<  HHBondsList.size() << " hh bonds." << endl;
  cout << "these are : " << endl;

  for (unsigned int i = 0; i < Protein::HHBondsList.size(); i++)
    {
      cout << Protein::HHBondsList[i].type << " " << Protein::HHBondsList[i].AA1 << " ";
      cout << Protein::HHBondsList[i].AA2  << endl;
    }
}

/* 
   does not directly set the hbond term energy itself.
   instead returns it.
*/
float Protein::computeHBondEnergyTerm()
{
  cout << "implement Protein::computeHBondEnergyTerm" << endl;
  return 0.0;
  #if 0
    float retval;
    for (i = 0; i < HHBondsList.size(); i++)
      {
	aa_1 = HHBondsList[i].first();
	aa_2 = HHBondsList[i].second();
	get N, C, o and H atom coordinates;
	compute angle;
	compute distance;
	if (both favorable)
	  retval += -1.0;
      }
    return retval;
  #endif
}

/*
  generates CA coords, if not generated yet.
  Then prints the pairwise distances, omitting those on the first and last
  loop vectors.
*/
void Protein::printPairwiseCADistances (const char * filename)
{
  if (cAlphaCoords == NULL) 
    {
      generateCAlphaCoords();
    }

  ofstream fout (filename);
  
  int numCAlphas     = lastAAIndexOfLastNonLoopVector_plusOne -
    firstAAIndexOfFirstNonLoopVector;

  for (int i = 0; i < numCAlphas; i++)
    {
      for (int j = 0; j < numCAlphas; j++)
	{
	  float vect[3];
	  VecSub (vect, cAlphaCoords[i], cAlphaCoords[j]);
	  fout << VecLength (vect) << " ";
	}
      fout << endl;
    }
  cout << "check " << filename << endl;
}

/*
  prints the coords of the protein.
*/
void Protein::printCACoords (const char * filename)
{
  if (cAlphaCoords == NULL) 
    {
      generateCAlphaCoords();
    }
  
  ofstream fout (filename);
  
  int numCAlphas     = lastAAIndexOfLastNonLoopVector_plusOne -
    firstAAIndexOfFirstNonLoopVector;
  
  for (int i = 0; i < numCAlphas; i++)
    {
      fout <<  cAlphaCoords[i] [0] << " ";
      fout <<  cAlphaCoords[i] [1] << " ";
      fout <<  cAlphaCoords[i] [2] << endl;
    }
  
  cout << "check " << filename << endl;
}
